
from epyk.core.html import graph


class SVG(object):
  def __init__(self, context):
    self.parent = context

  def new(self, width=(500, "px"), height=(300, "px")):
    """
    Description:
    ------------
    SVG stands for Scalable Vector Graphics.

    SVG defines vector-based graphics in XML format.

    Usage::

      svg = rptObj.ui.charts.svg.new(width=200)
      svg.add_text("I love SVG!", x=0, y=15, options={"fill": 'red'})

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/SVG
      https://www.w3schools.com/graphics/svg_intro.asp

    Attributes:
    ----------
    :param height: Optional. Integer for the component width
    :param width: Optional. Integer for the component height
    """
    if not isinstance(width, tuple):
      width = (width, "px")
    html_svg = graph.GraphSvg.SVG(self.parent.context.rptObj, width, height)
    self.parent.context.register(html_svg)
    return html_svg

  def line(self, x1=0, y1=None, x2=None, y2=None, width=(500, "px"), height=(300, "px"), options=None):
    """
    Description:
    ------------
    Entry point to the basic line definition in a SVG HTML Tag

    Usage::

      rptObj.ui.charts.svg.line(10, 30, 40, 69)

    Related Pages:

			https://www.w3schools.com/graphics/svg_line.asp

    Attributes:
    ----------
    :param x1: The x1 attribute defines the start of the line on the x-axis
    :param y1: The y1 attribute defines the start of the line on the y-axis
    :param x2: The x2 attribute defines the end of the line on the x-axis
    :param y2: The y2 attribute defines the end of the line on the y-axis
    :param width: Optional. Integer for the component width
    :param height: Optional. Integer for the component height
    :param options:
    """
    x2 = x2 or width[0]
    if y2 is None:
      y2 = 0 if height[1] == "%" else height[0] / 2
    if y1 is None:
      y1 = 0 if height[1] == "%" else height[0] / 2
    html_svg = graph.GraphSvg.SVG(self.parent.context.rptObj, width, height)
    html_svg.line(x1, y1, x2, y2)
    self.parent.context.register(html_svg)
    return html_svg

  def circle(self, x, y, r, width=(500, "px"), height=(300, "px"), options=None):
    """
    Description:
    ------------
    Entry point to the basic line definition in a SVG HTML Tag

    Usage::

      rptObj.ui.charts.svg.line(10, 30, 40, 69)

    Related Pages:

      https://www.w3schools.com/graphics/svg_line.asp

    Attributes:
    ----------
    :param x: The x attribute defines the start of the line on the x-axis
    :param y: The y attribute defines the start of the line on the y-axis
    :param r: The r attribute defines the radius
    :param width: Optional. Integer for the component width
    :param height: Optional. Integer for the component height
    :param options:
    """
    html_svg = graph.GraphSvg.SVG(self.parent.context.rptObj, width, height)
    html_svg.circle(x, y, r)
    self.parent.context.register(html_svg)
    return html_svg

  def arrow_right(self, x1=0, y1=None, x2=None, y2=None, size=10, width=(500, "px"), height=(300, "px"), options=None):
    """

    :param x1:
    :param y1:
    :param x2:
    :param y2:
    :param width:
    :param height:
    :param options:
    """
    line = self.line(x1, y1, x2 or width[0]-size, y2, width, height, options)
    defs = line.defs()
    m = defs.marker("arrow", "0 0 10 10", 5, 5)
    m.arrow(size)
    line[0].marker_end(m.url)
    return line

  def arrow_left(self, x1=0, y1=None, x2=None, y2=None, size=10, width=(500, "px"), height=(300, "px"), options=None):
    """

    Usage::

      rptObj.ui.charts.svg.arrow_left()

    :param x1:
    :param y1:
    :param x2:
    :param y2:
    :param width:
    :param height:
    :param options:
    """
    line = self.line(x1+size, y1, x2, y2, width, height, options)
    defs = line.defs()
    m = defs.marker("arrow", "0 0 10 10", 5, 5)
    m.arrow(size).orient("auto-start-reverse")
    line[0].marker_start(m.url)
    return line

  def ellipse(self, cx, cy, rx, ry, width=(500, "px"), height=(300, "px"), options=None):
    """
    Description:
    ------------
    SVG Ellipse - <ellipse>

    Usage::

      rptObj.ui.charts.svg.ellipse(100, 100, 40, 69)

    Related Pages:

			https://www.w3schools.com/graphics/svg_ellipse.asp

    Attributes:
    ----------
    :param cx: The cx attribute defines the x coordinate of the center of the ellipse
    :param cy: The cy attribute defines the y coordinate of the center of the ellipse
    :param rx: The rx attribute defines the horizontal radius
    :param ry: The ry attribute defines the vertical radius
    :param width: Optional. Integer for the component width
    :param height: Optional. Integer for the component height
    :param options:
    """
    html_svg = graph.GraphSvg.SVG(self.parent.context.rptObj, width, height)
    html_svg.ellipse(cx, cy, rx, ry)
    self.parent.context.register(html_svg)
    return html_svg

  def polyline(self, points, width=(500, "px"), height=(300, "px"), options=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.charts.svg.polyline([(15, 80), (29, 50), (43, 60), (57, 30), (71, 40), (85, 15)])

    Related Pages:

			https://www.w3schools.com/graphics/svg_polyline.asp

    Attributes:
    ----------
    :param points: The points attribute defines the list of points (pairs of x and y coordinates) required to draw the polyline
    :param height:
    :param width:
    :param options:
    """
    html_svg = graph.GraphSvg.SVG(self.parent.context.rptObj, width, height)
    html_svg.polyline(points)
    self.parent.context.register(html_svg)
    return html_svg

  def polygone(self, points, width=(500, "px"), height=(300, "px"), options=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.charts.svg.polygone([(15, 80), (29, 50), (43, 60), (57, 30), (71, 40), (85, 15)])

    Related Pages:

			https://www.w3schools.com/graphics/tryit.asp?filename=trysvg_polygon

    Attributes:
    ----------
    :param points: The points attribute defines the list of points (pairs of x and y coordinates) required to draw the polyline
    :param height:
    :param width:
    :param options:
    """
    html_svg = graph.GraphSvg.SVG(self.parent.context.rptObj, width, height)
    html_svg.polygon(points)
    self.parent.context.register(html_svg)
    return html_svg

  def triangle(self, point1, point2=None, point3=None, fill='None', width=(500, "px"), height=(300, "px"), options=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.charts.svg.triangle((50, 100))

    Related Pages:

			https://www.w3schools.com/graphics/svg_polyline.asp

    Attributes:
    ----------
    :param point1:
    :param point2:
    :param point3:
    :param fill:
    :param width:
    :param height:
    :param options:
    """
    if point2 is None:
      point2 = (point1[1]/2, point1[0])
    if point3 is None:
      point3 = (point1[1], point1[1])
    tri = graph.GraphSvg.SVG(self.parent.context.rptObj, width, height)
    tri.triangle([point1, point2, point3, point1], fill=fill, options=options)
    self.parent.context.register(tri)
    return tri

  def axes(self, size=10, width=(500, "px"), height=(300, "px")):
    """
    Description:
    ------------

    Usage::

      svg = rptObj.ui.charts.svg.axes()
      m = svg.defs().marker("circle", "0 0 10 10", 5, 5)
      m.circle(5, 5, 5, 'red')
      m.markerWidth(10).markerHeight(10)
      p = svg.path(0, 0, from_origin=True).line_to(50, 100).horizontal_line_to(300).line_to(400, 200)
      p.markers(m.url)

    Attributes:
    ----------
    :param size:
    :param width:
    :param height:
    """
    svg = graph.GraphSvg.SVG(self.parent.context.rptObj, width, height)
    svg.origine = (size, height[0]-size)
    defs = svg.defs()
    m = defs.marker("arrow", "0 0 10 10", 5, 5)
    m.arrow().orient("auto-start-reverse")
    m.markerWidth(size).markerHeight(size)
    pl = svg.polyline([(size, size), (size, height[0]-size), (width[0]-size, height[0]-size)]).css({'stroke': "black"})
    pl.marker_start(m.url)
    pl.marker_end(m.url)
    self.parent.context.register(svg)
    return svg

  def rectangle(self, x, y, width=(500, "px"), height=(300, "px"), fill=None, rx=0, ry=0):
    """

    :param x:
    :param y:
    :param width:
    :param height:
    :param fill:
    :param rx:
    :param ry:
    """
    rect = graph.GraphSvg.SVG(self.parent.context.rptObj, (900, "px"), (200, "px"))
    rect.rect(x, y, width, height, fill, rx=rx, ry=ry)
    self.parent.context.register(rect)
    return rect

  def heart(self, w, h, fill='none', width=(500, "px"), height=(300, "px")):
    """

    c = rptObj.ui.charts.svg.heart(w=50, h=100, fill="pink")
    c[0].transform("transform", "rotate", "0 100 100", "360 100 10")

    :param w:
    :param h:
    :param fill:
    :param width:
    :param height:
    """
    svg = graph.GraphSvg.SVG(self.parent.context.rptObj, width, height)
    svg.polygon([(w, w), (3/4 * h, 3/4 * h), (h, w), (h, 3/4 * w), (9/10 * h, 6/10 * w), (8.5/10 * h, 6/10 * w), (3/4 * h, 3/4 * w),
                 (2/3 * h, 6/10 * w), (6/10 * h, 6/10 * w), (w, 3/4 * w), (w, w)], fill=fill)
    self.parent.context.register(svg)
    return svg

  def star(self, fill='none', width=(500, "px"), height=(300, "px")):
    """

    https://codepen.io/susanwinters/pen/WxbRJK

    :param w:
    :param h:
    :param fill:
    :param width:
    :param height:
    """
    svg = graph.GraphSvg.SVG(self.parent.context.rptObj, width, height)
    svg.polygon([(100, 10), (40, 180), (190, 60), (10, 60), (160, 180)], fill=fill)
    svg[-1].css({"stroke": 'none'})
    self.parent.context.register(svg)
    return svg

  def path(self, x=0, y=0, fill='none', origin=False, bespoke_path=None):
    """

    :param x:
    :param y:
    :param fill:
    :param origin:
    :param bespoke_path:
    :return:
    """
    path = graph.GraphSvg.Path(self.parent.context.rptObj, x, y, fill, origin, bespoke_path)
    self.parent.context.register(path)
    return path


import json

from epyk.core.js import JsUtils


class JsSwitch(object):
  """

  Documentation:
    - https://www.w3schools.com/js/js_switch.asp

  """

  def __init__(self, jsObj, context=None):
    """

    :param jsCondition:
    :param jsFncs:
    :param context:
    """
    self._context = context
    self.__jsObj = jsObj
    self.__js, self.__default = [], None

  def case(self, value, jsFncs):
    """

    :param jsFncs:
    :param isPyData:
    :param jsFncVal:
    :return:
    """
    self.__js.append((value, jsFncs))
    return self

  def default_(self, jsFncs):
    """

    :param jsFncs:
    :return:
    """
    self.__default = jsFncs
    return self

  def toStr(self):
    """

    :return:
    """
    strData = []
    for val, jsFncs in self.__js:
      strData.append("case %s: {%s; break}" % (json.dumps(val), ";".join(map(lambda x: str(x),  jsFncs))))
    if self.__default is not None:
      strData.append("default: {%s}" % ";".join(self.__default))
    self.__jsFncs, self.__default = [], None # empty the stack
    return "switch (%s){%s}" % (self.__jsObj.varName, "".join(strData))



"""
Module dedicated to perform the data transformation for the Billboard / C3 charts

Related Pages:

		https://naver.github.io/billboard.js/
Related Pages:

		https://c3js.org/
"""

from epyk.core.js import JsUtils


class JsChartBillboardLinks(object):
  def __init__(self, data, js_src, data_schema=None, profile=False):
    self._js_src, self._data_schema, self.profile, self._data = js_src, data_schema, profile, data

  def __register_records_fnc(self, fnc_name, fnc_def, fnc_pmts=None, profile=False):
    """
    This function will attach to the report object only the javascript functions used during the report

    :param fnc_name: A String with the Javascript function name to be defined
    :param fnc_def: A String with the Javascript function content
    :param fnc_pmts: A list of parameters
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    fnc_pmts = ["data"] + (fnc_pmts or [])
    if not fnc_name in self._js_src.get('js', {}).get('functions', {}):
      if profile:
        content = "var result = []; %s;return result" % JsUtils.cleanFncs(fnc_def)
      else:
        content = "var result = []; %s;return result" % JsUtils.cleanFncs(fnc_def)
      self._js_src.setdefault('js', {}).setdefault('functions', {})[fnc_name] = {'content': content, 'pmt': fnc_pmts}

  def line(self, y_columns, x_axis, profile=False):
    """
    Data Conversion to a Line chart object.

    This entry point will register the Javascript function and then use it on the javascript side using the parameters
    passed in this function. Some basic controls are done on the Python side during the transpilation

    Example

    :param y_columns: A python list with the columns in the records
    :param x_axis: A string with a column name from the records
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsBillDefault.__name__
    x_axis = JsUtils.jsConvertData(x_axis, None)
    self.__register_records_fnc(fnc_name, JsBillDefault.content, list(JsBillDefault.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s, %s)" % (fnc_name, y_columns, x_axis)

  def pie(self, y_columns, x_axis, profile=False):
    """
    Data Conversion to a Pie chart object.

    This entry point will register the Javascript function and then use it on the javascript side using the parameters
    passed in this function. Some basic controls are done on the Python side during the transpilation

    Example

    :param y_columns: A python list with the columns in the records
    :param x_axis: A string with a column name from the records
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsBillPie.__name__
    x_axis = JsUtils.jsConvertData(x_axis, None)
    self.__register_records_fnc(fnc_name, JsBillPie.content, list(JsBillPie.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s, %s)" % (fnc_name, y_columns, x_axis)

  def scatter(self, y_columns, x_axis, profile=False):
    """
    Data Conversion to a Scatter chart object.

    This entry point will register the Javascript function and then use it on the javascript side using the parameters
    passed in this function. Some basic controls are done on the Python side during the transpilation

    Example

    :param y_columns: A python list with the columns in the records
    :param x_axis: A string with a column name from the records
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsBillScatter.__name__
    x_axis = JsUtils.jsConvertData(x_axis, None)
    self.__register_records_fnc(fnc_name, JsBillScatter.content, list(JsBillScatter.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s, %s)" % (fnc_name, y_columns, x_axis)


# --------------------------------------------------------------------------------------------------------------------
#                               DATA TRANSFORMATION JAVASCRIPT DEFINITION
# --------------------------------------------------------------------------------------------------------------------
class JsBillDefault(object):
  pmts = ("seriesNames", "xAxis")
  content = '''
    var temp = {}; var labels = []; var uniqLabels = {};
    seriesNames.forEach(function(series){temp[series] = {}});
    data.forEach(function(rec){ 
      seriesNames.forEach(function(name){
        if(rec[name] !== undefined){
          if (!(rec[xAxis] in uniqLabels)){labels.push(rec[xAxis]); uniqLabels[rec[xAxis]] = true};
          temp[name][rec[xAxis]] = rec[name]}})});
    result = []; result.push(['x'].concat(labels));
    seriesNames.forEach(function(series){
      dataSet = [series];
      labels.forEach(function(x){
        if(temp[series][x] == undefined){dataSet.push(null)} else {dataSet.push(temp[series][x])}
      }); result.push(dataSet)})
    '''


class JsBillScatter(object):
  pmts = ("seriesNames", "xAxis")
  content = '''
    var tempVal = {}; var tempX = {}; var labels = []; 
    seriesNames.forEach(function(series){tempVal[series] = []; tempX[series +"_x"] = []});
    data.forEach(function(rec){ 
      seriesNames.forEach(function(name){
        if(rec[name] !== undefined){
          if(!(rec[xAxis] in tempVal[name])){tempVal[name] = [name, rec[name]]; tempX[name +"_x"] = [name +"_x", rec[xAxis]]}
          else {tempVal[name].push(rec[name]); tempX[name +"_x"].push(rec[xAxis])}}})});
    result = {'columns': [], 'xs': {}};
    seriesNames.forEach(function(series){
      result.columns.push(tempVal[series]); result.columns.push(tempX[series +"_x"]); result.xs[series] = series +"_x"})
    '''


class JsBillPie(object):
  pmts = ("seriesNames", "xAxis")
  content = '''
    var temp = {}; var labels = {};
    data.forEach(function(rec){ 
      if(!(rec[xAxis] in temp)){temp[rec[xAxis]] = {}};
      seriesNames.forEach(function(name){
        labels[name] = true; 
        if(rec[name] !== undefined){
          if(!(name in temp[rec[xAxis]])){temp[rec[xAxis]][name] = rec[name]} else{temp[rec[xAxis]][name] += rec[name]}}})});
    result = []; result.push(['x'].concat(labels)); var labels = Object.keys(labels);
    for(var series in temp){
      var values = [series];
      labels.forEach(function(label){
        if(temp[series][label] !== undefined){values.push(temp[series][label])} else{values.push(null)}});
      result.push(values)}
    '''




from epyk.core.html.options import Options


class OptionsPanelPoints(object):

  def __init__(self, src, options):
    self.src = src
    self.__background_color = options.get("background-color", src.theme.success[1])
    self.__div_css = options.get("div_css", {})
    self.__selected = options.get("selected", 0)

  @property
  def background_color(self):
    return self.__background_color

  @background_color.setter
  def background_color(self, val):
    self.__background_color = val

  @property
  def div_css(self):
    return self.__div_css

  @div_css.setter
  def div_css(self, css):
    self.__div_css = css

  @property
  def selected(self):
    return self.__selected

  @selected.setter
  def selected(self, num):
    self.__selected = num


class OptionPanelSliding(Options):

  @property
  def expanded(self):
    """
    Description:
    ------------

    :return:
    """
    return self.get(True)

  @expanded.setter
  def expanded(self, boool):
    self.set(boool)

  @property
  def icon_expanded(self):
    """
    Description:
    ------------

    :return:
    """
    return self.get("fas fa-caret-down")

  @icon_expanded.setter
  def icon_expanded(self, icon):
    self.set(icon)

  @property
  def icon_closed(self):
    """
    Description:
    ------------

    :return:
    """
    return self.get("fas fa-caret-up")

  @icon_closed.setter
  def icon_closed(self, icon):
    self.set(icon)


class OptionPanelTabs(Options):

  @property
  def css_tab(self):
    """
    Description:
    ------------
    The default CSS style for the tabs.
    This must be changed before adding components
    """
    dflt = {'display': 'inline-block', 'text-align': 'center', 'cursor': 'pointer', 'margin': '0 2px 5px 0', "border-bottom": "1px solid white"}
    return self.get(dflt)

  @css_tab.setter
  def css_tab(self, attrs):
    self.set(attrs)

  @property
  def css_tab_clicked(self):
    """
    Description:
    ------------
    The default CSS style for the clicked tab.
    This must be changed before adding components
    """
    dflt = {"border-bottom": "1px solid %s" % self._report.theme.success[1]}
    return self.get(dflt)

  @css_tab_clicked.setter
  def css_tab_clicked(self, attrs):
    self.set(attrs)

  def tab_style(self, name, css_style=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param name:
    :param css_style:
    """
    if not 'tab_style' in self._attrs:
      self._attrs['tab_style'] = {}
    css = dict(self.css_tab)
    if css_style is not None:
      css.update(css_style)
      self._attrs['tab_style'][name] = css
    return self._attrs['tab_style'].get(name, css)

  def tab_clicked_style(self, name, css_style=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param name:
    :param css_style:
    """
    if not 'tab_style_clicked' in self._attrs:
      self._attrs['tab_style_clicked'] = {}
    css = dict(self.css_tab_clicked)
    if css_style is not None:
      css.update(css_style)
      self._attrs['tab_style_clicked'][name] = css
    return self._attrs['tab_style_clicked'].get(name, css)

  def tab_not_clicked_style(self, name=None, css_style=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param name:
    :param css_style:
    """
    tab_clicked_style = self.tab_clicked_style(name)
    tab_style = self.tab_style(name)
    tab_not_clicked = {k: tab_style[k] if k in tab_style else 'none' for k, v in tab_clicked_style.items()}
    if css_style is not None:
     tab_not_clicked.update(css_style)
    return tab_not_clicked


class OptionPanelTable(Options):

  @property
  def header(self):
    """
    Description:
    ------------
    """
    return self.get(True)

  @header.setter
  def header(self, bool):
    self.set(bool)

  @property
  def cell_align(self):
    """
    Description:
    ------------
    """
    return self.get()

  @cell_align.setter
  def cell_align(self, value):
    self.set(value)


class OptionsDiv(Options):

  @property
  def inline(self):
    """
    Description:
    ------------
    """
    return self.get(False)

  @inline.setter
  def inline(self, bool):
    self.set(bool)


class OptionDrawer(Options):

  @property
  def side(self):
    """
    Description:
    ------------
    """
    return self.get("left")

  @side.setter
  def side(self, value):
    self.set(value)

  @property
  def width(self):
    """
    Description:
    ------------
    """
    return self.get("200px")

  @width.setter
  def width(self, integer):
    if isinstance(integer, int):
      integer = "%spx" % integer
    self.set(integer)


class OptionsStepper(Options):

  def __add_colors(self, type, colors):
    """
    Description:
    ------------
    Set the colors for a step state

    Attributes:
    ----------
    :param type: String. The state
    :param colors: List or Dictionary. The color definition
    """
    if not 'colors' in self._report._jsStyles:
      self._report._jsStyles['colors'] = {}
    if not isinstance(colors[0], dict):
      s = 100 / (len(colors) - 1)
      tmp_colors = []
      for i, c in enumerate(colors):
        tmp_colors.append({"color": c, 'offset': "%s%%" % int(i * s)})
      colors = tmp_colors
    self._report._jsStyles['colors'][type] = colors

  @property
  def success(self):
    """
    Description:
    ------------
    Add the success colors
    """
    return self._report._jsStyles.get('colors', {}).get('success', [])

  @success.setter
  def success(self, colors):
    self.__add_colors('success', colors)

  @property
  def error(self):
    """
    Description:
    ------------

    """
    return self._report._jsStyles['colors'].get('error', [])

  @error.setter
  def error(self, colors):
    self.__add_colors('error', colors)

  @property
  def pending(self):
    """
    Description:
    ------------

    """
    return self._report._jsStyles['colors'].get('pending', [])

  @pending.setter
  def pending(self, colors):
    self.__add_colors('pending', colors)

  @property
  def waiting(self):
    """
    Description:
    ------------
    The list of
    """
    return self._report._jsStyles['colors'].get('waiting', [])

  @waiting.setter
  def waiting(self, colors):
    self.__add_colors('waiting', colors)

  @property
  def blink(self):
    """
    Description:
    ------------

    """
    return self._report._jsStyles.get('colors', {}).get('blink')

  @blink.setter
  def blink(self, colors):
    self.__add_colors('blink', colors)

  @property
  def circle_factor(self):
    """
    Description:
    ------------

    """
    return self._config_get(False)

  @circle_factor.setter
  def circle_factor(self, bool):
    self._config(bool)

  @property
  def width(self):
    """
    Description:
    ------------

    """
    return self._config_get(100)

  @width.setter
  def width(self, num):
    self._config(num)

  @property
  def shape(self):
    """
    Description:
    ------------

    """
    return self._config_get('arrow')

  @shape.setter
  def shape(self, value):
    if not value in ('arrow', 'triangle', 'rectangle', 'circle'):
      raise Exception("This shape is not yet available, please create a request for adding it")

    self._config(value)

  @property
  def opacities(self):
    """
    Description:
    ------------

    """
    return self._config_get([])

  @opacities.setter
  def opacities(self, values):
    self._config(values)

  @property
  def height(self):
    """
    Description:
    ------------

    """
    return self._config_get(50)

  @height.setter
  def height(self, num):
    self._config(num)

  @property
  def line(self):
    """
    Description:
    ------------

    """
    return self._config_get(False)

  @line.setter
  def line(self, attrs):
    if attrs is True:
      attrs = {"stroke": 'grey', 'stroke-width': 2}
    self._config(attrs)

  @property
  def backgrounds(self):
    """
    Description:
    ------------

    """
    return self._config_get({})

  @backgrounds.setter
  def backgrounds(self, color):
    self._config(color)

  @property
  def svg_style(self):
    """
    Description:
    ------------

    """
    return self._config_get({})

  @svg_style.setter
  def svg_style(self, css):
    if not 'svg_style' in self._report._jsStyles:
      self._report._jsStyles['svg_style'] = {}
    self._report._jsStyles['svg_style'].update(css)

  @property
  def text_color(self):
    """
    Description:
    ------------

    """
    return self._config_get('white')

  @text_color.setter
  def text_color(self, colors):
    self._config(colors)

  @property
  def text_style(self):
    """
    Description:
    ------------

    """
    return self._config_get({})

  @text_style.setter
  def text_style(self, css):
    if not 'title_style' in self._report._jsStyles:
      self._report._jsStyles['text_style'] = {}
    self._report._jsStyles['text_style'].update(css)




from epyk.core.html.options import Options


class OptionsTree(object):

  def __init__(self, src, options):
    self._icon_open, self._icon_close, self._expanded = "fas fa-folder-open", "fas fa-folder", True
    self.src = src

  @property
  def icon_open(self):
    """

    :return:
    """
    return self._icon_open

  @property
  def icon_close(self):
    """

    :return:
    """
    return self._icon_close

  @property
  def expanded(self):
    """

    :return:
    """
    return self._expanded

  @expanded.setter
  def expanded(self, bool):
    self._expanded = bool
    self.src.items = None
    self.src.set(self.src, self.src.val)

  @icon_open.setter
  def icon_open(self, icon):
    self._icon_open = icon
    self.src.items = []
    self.src.set(self.src, self.src.val)

  @icon_close.setter
  def icon_close(self, icon):
    self._icon_close = icon
    self.src.items = []
    self.src.set(self.src, self.src.val)


class OptDropDown(Options):

  @property
  def width(self):
    """
    Description:
    ------------
    """
    return self._report._jsStyles.get('autoFocus', False)

  @width.setter
  def width(self, value):
    if isinstance(value, int):
      value = "%spx" % value
    #
    self._report._jsStyles["a"]["width"] = value
    self._report._jsStyles["ul"]["left"] = value
    return self





from epyk.core.data import DataClass


class Scroller(DataClass):

  def activate(self):
    """
    Description:
    -----------
    Enable and configure the Scroller extension for DataTables

    Related Pages:

			https://datatables.net/reference/option/scroller
    """
    self.loadingIndicator = True
    return self

  @property
  def boundaryScale(self):
    """
    Description:
    -----------
    Set the point at which new data will be loaded and drawn.

    Related Pages:

			https://datatables.net/reference/option/scroller.boundaryScale
    """
    return self._attrs["boundaryScale"]

  @boundaryScale.setter
  def boundaryScale(self, val):
    self._attrs["boundaryScale"] = val

  @property
  def displayBuffer(self):
    """
    Description:
    -----------
    The amount of data that Scroller should pre-buffer to ensure smooth scrolling.

    Related Pages:

			https://datatables.net/reference/option/scroller.displayBuffer
    """
    return self._attrs["displayBuffer"]

  @displayBuffer.setter
  def displayBuffer(self, val):
    self._attrs["displayBuffer"] = val

  @property
  def loadingIndicator(self):
    """
    Description:
    -----------
    Display a loading message while Scroller is loading additional data.

    Related Pages:

			https://datatables.net/reference/option/scroller.loadingIndicator
    """
    return self._attrs["loadingIndicator"]

  @loadingIndicator.setter
  def loadingIndicator(self, val):
    self._attrs["loadingIndicator"] = val

  @property
  def rowHeight(self):
    """
    Description:
    -----------
    Set the row height, or how the row height is calculated.

    Related Pages:

			https://datatables.net/reference/option/scroller.rowHeight
    """
    return self._attrs["rowHeight"]

  @rowHeight.setter
  def rowHeight(self, val):
    self._attrs["rowHeight"] = val

  @property
  def serverWait(self):
    """
    Description:
    -----------
    Time delay before new data is requested when server-side processing is enabled.

    Related Pages:

			https://datatables.net/reference/option/scroller.serverWait
    """
    return self._attrs["serverWait"]

  @serverWait.setter
  def serverWait(self, val):
    self._attrs["serverWait"] = val




from epyk.core import html


class Layouts(object):
  def __init__(self, context):
    self.context = context

  def new_line(self, count=1, profile=None):
    """
    Description:
    ------------
    Wrapper around the Br html tag.

    The <br> tag inserts a single line break.

    Usage::

      rptObj.ui.layouts.new_line(10)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlOthers.Newline`

    Related Pages:

			https://www.w3schools.com/tags/tag_br.asp

    Attributes:
    ----------
    :param count: Optional, The number of empty line to put. Default 1
    :param profile: Optional, Activate the profiler
    """
    html_new_line = html.HtmlOthers.Newline(self.context.rptObj, count, profile=profile)
    self.context.register(html_new_line)
    return html_new_line

  def hr(self, count=1, color=None, background_color=None, height=(None, 'px'), align=None, profile=None):
    """
    Description:
    ------------
    Wrapper around the HT html tag.

    The <hr> tag defines a thematic break in an HTML page (e.g. a shift of topic).

    Usage::

      rptObj.ui.layouts.hr(10)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlOthers.Hr`

    Related Pages:

			https://www.w3schools.com/tags/tag_hr.asp

    Attributes:
    ----------
    :param count: The number of HR tag to be added
    :param color: Optional. The color code for the font
    :param background_color: Optional. The component background color
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param align: Optional. The content position. Values (left, right, center). Default center
    :param profile: Optional. A flag to set the component performance storage
    """
    hr_html = self.context.rptObj.ui.div()
    for _ in range(count):
      hr_item = html.HtmlOthers.Hr(self.context.rptObj, background_color, height, align, profile)
      self.context.register(hr_item)
      hr_html += hr_item
    return hr_html

  def col(self, htmlObjs=None, position='middle', width=(100, '%'), height=(None, 'px'), align=None, helper=None, profile=None):
    """
    Description:
    ------------
    Python wrapper for a column of HTML elements from Bootstrap

    This component is a container and it is used to display multiple Ares components in column.
    You can first add a component in the data list then add the + operator to add more.

    Usage::

      rptObj.ui.layouts.col([
      rptObj.ui.text("test C"),
      rptObj.ui.text("test D"),
      ])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Col`

    Related Pages:

			https://getbootstrap.com/docs/4.0/layout/grid/
    https://www.alsacreations.com/tuto/lire/1493-css3-flexbox-layout-module.html

    Attributes:
    ----------
    :param htmlObjs:
    :param position:
    :param width:
    :param height:
    :param align:
    :param helper:
    :param profile:
    """
    html_col = html.HtmlContainer.Col(self.context.rptObj, htmlObjs, position, width, height, align, helper, profile)
    self.context.register(html_col)
    return html_col

  def row(self, htmlObjs=None, position='middle', width=(100, '%'), height=(None, 'px'), align=None, helper=None, profile=None):
    """
    Description:
    ------------
    Python wrapper for a column of HTML elements from Bootstrap

    This component is a container and it is used to display multiple Ares components in column.
    You can first add a component in the data list then add the + operator to add more.

    Usage::

      row = rptObj.ui.layouts.row()
      row += rptObj.ui.layouts.col([
      rptObj.ui.text("test A"),
      rptObj.ui.text("test B"),
      ])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Row`

    Related Pages:

			https://getbootstrap.com/docs/4.0/layout/grid/
    https://www.alsacreations.com/tuto/lire/1493-css3-flexbox-layout-module.html

    Attributes:
    ----------
    :param htmlObjs:
    :param position:
    :param width:
    :param height:
    :param align:
    :param helper:
    :param profile:
    """
    html_col = html.HtmlContainer.Row(self.context.rptObj, htmlObjs, position, width, height, align, helper, profile)
    self.context.register(html_col)
    return html_col

  def table(self, htmlObjs=None, width=(100, '%'), height=(None, 'px'), helper=None, options=None, profile=None):
    """
    Description:
    ------------
    table layout for HTML components

    Usage::

      row = rptObj.ui.layouts.table()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Table`

    Attributes:
    ----------
    :param htmlObjs:
    :param width:
    :param height:
    :param helper:
    :param options:
    :param profile:
    """
    html_row = html.HtmlContainer.Table(self.context.rptObj, htmlObjs, width, height, helper, options, profile)
    self.context.register(html_row)
    return html_row

  def grid(self, rows=None, width=(100, '%'), height=(None, 'px'), align=None, helper=None, profile=None):
    """
    Description:
    ------------
    Python wrapper to the HTML Bootstrap Grid

    Usage::

      gr = rptObj.ui.layouts.grid()
      gr += [rptObj.ui.text("test %s" % i) for i in range(5)]

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Grid`

    Related Pages:

			https://getbootstrap.com/docs/4.0/layout/grid/

    Attributes:
    ----------
    :param rows:
    :param width:
    :param height:
    :param align:
    :param helper:
    :param profile:
    """
    html_grid = html.HtmlContainer.Grid(self.context.rptObj, rows, width, height, align, helper, profile)
    self.context.register(html_grid)
    return html_grid

  def panel(self, htmlObjs=None, title=None, color=None, width=(100, "%"), height=(None, "px"), htmlCode=None,
            helper=None, options=None, profile=False):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Panel`

    :param htmlObjs:
    :param title:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    """
    if htmlObjs is not None and not isinstance(htmlObjs, list):
      htmlObjs = [htmlObjs]
    html_panel = html.HtmlContainer.Panel(self.context.rptObj, htmlObjs or [], title, color, width, height, htmlCode, helper, options, profile)
    self.context.register(html_panel)
    return html_panel

  def div(self, htmlObjs=None, label=None, color=None, width=(100, "%"), icon=None, height=(None, "px"), editable=False,
          align='left', padding=None, htmlCode=None, tag='div', helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      div = rptObj.ui.div([html])
      div += html_2

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`

    Related Pages:

			https://www.w3schools.com/tags/tag_div.asp

    Attributes:
    ----------
    :param htmlObjs:
    :param label:
    :param color:
    :param width:
    :param icon:
    :param height:
    :param editable:
    :param align:
    :param padding:
    :param htmlCode:
    :param tag:
    :param profile:
    """
    if htmlObjs is not None and not isinstance(htmlObjs, list):
      htmlObjs = [htmlObjs]
    html_div = html.HtmlContainer.Div(self.context.rptObj, htmlObjs or [], label, color, width, icon, height,
                                      editable, align, padding, htmlCode, tag, helper, options or {}, profile)
    self.context.register(html_div)
    return html_div

  def popup(self, htmlObj=None, title=None, color=None, width=(100, '%'), height=(None, 'px'),
            withBackground=True, draggable=False, margin=10, profile=None):
    """
    Description:
    ------------

    Usage::

      popup = report.popup(report.title('Test'), color="red")
      popup + report.paragraph('Test')

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlPopup.Popup`

    Related Pages:

			https://www.w3schools.com/tags/tag_div.asp

    Attributes:
    ----------
    :param htmlObj:
    :param title:
    :param color:
    :param width:
    :param height:
    :param withBackground:
    :param draggable:
    :param margin:
    :param profile:
    :rtype: html.HtmlPopup.Popup
    :return:
    """
    return self.context.register(html.HtmlPopup.Popup(self.context.rptObj, htmlObj, title, color, width, height,
                                                      withBackground, draggable, margin, profile))

  def iframe(self, url, width=(100, "%"), height=(100, "%"), helper=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.layouts.iframe("http://www.google.com")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.IFrame`

    Attributes:
    ----------
    :param url:
    :param width:
    :param height:
    :param helper:
    :param profile:
    """
    html_frame = html.HtmlContainer.IFrame(self.context.rptObj, url, width, height, helper, profile)
    self.context.register(html_frame)
    return html_frame

  def dialogs(self, record=None, width=(100, "%"), height=(200, "px"), helper=None, profile=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Dialog`

    :param record:
    :param width:
    :param height:
    :param helper:
    :param profile:
    :return:
    """
    html_dialog = html.HtmlContainer.Dialog(self.context.rptObj, record, width, height, helper, profile)
    self.context.register(html_dialog)
    return html_dialog

  def icons(self, icon_names=None, width=(100, "%"), height=(None, "px"), htmlCode=None, helper=None, profile=None):
    """
    Description:
    ------------

    Usage::

      menu = rptObj.ui.layouts.icons(["fas fa-bell", "fas fa-calendar-check"])
      menu.icon.click([menu.icon.dom.css({"color": 'red'})])
      menu[0].click([menu[0].dom.css({"color": 'red'})])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.IconsMenu`

    Attributes:
    ----------
    :param icon_names:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param profile:
    """
    icon_names = icon_names or None
    if not isinstance(icon_names, list):
      icon_names = [icon_names]
    html_icon = html.HtmlContainer.IconsMenu(icon_names, self.context.rptObj, width, height, htmlCode, helper, profile)
    self.context.register(html_icon)
    return html_icon

  def form(self, htmlObj=None, helper=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Form`

    Attributes:
    ----------
    :param htmlObj:
    :param helper:
    :return:
    """
    form = html.HtmlContainer.Form(self.context.rptObj, htmlObj, helper)
    self.context.register(form)
    return form

  def header(self, htmlObjs=None, width=(100, "%"),  height=(None, "px"), htmlCode=None, helper=None, options=None, profile=None):
      """
      Description:
      ------------
      The HTML <header> element represents introductory content, typically a group of introductory or navigational aids.
      It may contain some heading elements but also a logo, a search form, an author name, and other elements

      Usage::

        div = rptObj.ui.header([html])
        div += html_2

      Underlying HTML Objects:

        - :class:`epyk.core.html.HtmlContainer.Header`

      Related Pages:

  			https://developer.mozilla.org/en-US/docs/Web/HTML/Element/header

      Attributes:
      ----------
      :param htmlObjs:
      :param width:
      :param height:
      :param htmlCode:
      :param tag:
      :param profile:
      """
      if htmlObjs is not None and not isinstance(htmlObjs, list):
        htmlObjs = [htmlObjs]
      html_obj = html.HtmlContainer.Header(self.context.rptObj, htmlObjs or [], width, height, htmlCode, helper, options or {}, profile)
      self.context.register(html_obj)
      return html_obj

  def section(self, htmlObjs=None, width=(100, "%"),  height=(None, "px"), htmlCode=None, helper=None, options=None, profile=None):
      """
      Description:
      ------------
      The <section> tag defines sections in a document, such as chapters, headers, footers, or any other sections of the document.

      Usage::

        div = rptObj.ui.header([html])
        div += html_2

      Underlying HTML Objects:

        - :class:`epyk.core.html.HtmlContainer.Header`

      Related Pages:

  			https://www.w3schools.com/tags/tag_section.asp

      Attributes:
      ----------
      :param htmlObjs:
      :param width:
      :param height:
      :param htmlCode:
      :param tag:
      :param profile:
      """
      if htmlObjs is not None and not isinstance(htmlObjs, list):
        htmlObjs = [htmlObjs]
      html_obj = html.HtmlContainer.Section(self.context.rptObj, htmlObjs or [], width, height, htmlCode, helper, options or {}, profile)
      self.context.register(html_obj)
      return html_obj




import json

from epyk.core.js.primitives import JsArray
from epyk.core.js.primitives import JsObject
from epyk.core.js.primitives import JsBoolean
from epyk.core.js.primitives import JsNumber
from epyk.core.js.primitives import JsString

from epyk.core.js.objects import JsNodeDom

from epyk.core.js.packages.JsVis import VisDataSet, VisDataView

from epyk.core.js.fncs import JsFncs
from epyk.core.js import JsUtils


class DataLoop(object):
  """
  Data Class used for all the loop and map in the Javascript side.
  This will get the below attributes

  val   : The current value in the loop
  index : The index item
  arr   : The full array (only available in case of arrays, map, filter, every  )
  """
  val, index, arr = JsObject.JsObject("value"), JsNumber.JsNumber("index", isPyData=False), JsArray.JsArray("arr")


class DataReduce(object):
  """

  rVal  :
  val   :
  index :
  """
  rVal, val, index = JsObject.JsObject("r"), JsNumber.JsNumber("o", isPyData=False), JsNumber.JsNumber("i", isPyData=False)


class DataSort(object):
  """

  """


class DataEach(object):
  """
  Data Class for the Jquery each loop

  index : index
  data  : element
  """
  index, data = JsNumber.JsNumber("index", isPyData=False), JsObject.JsObject("data", isPyData=False)


class DataAll(object):
  """
  Data Class for the Jquery each loop

  index : index
  data  : elt
  """
  index, element = JsNumber.JsNumber("index", isPyData=False), JsNodeDom.JsDoms.get(varName="elt")


class ContainerData(object):
  def __init__(self, report, schema):
    self._report, self._schema = report, schema

  @property
  def fnc(self):
    """
    All the predefined transformation functions.
    """
    return JsFncs.FncOnRecords(self._report._props, self._schema)

  @property
  def to(self):
    """
    All the possible object transformation to deal with external packages
    """
    return JsFncs.FncToObject(self._report._props, self._schema)

  @property
  def agg(self):
    """
    Property to the data aggregator functions
    The aggregator will create a new record with different column names
    """
    return JsFncs.FncRoAggRec(self._report._props, self._schema)

  @property
  def filter(self):
    """

    """
    return JsFncs.FncFiltere(self, self._report._props, self._schema)


class RawData(object):
  def __init__(self, report, records=None, profile=False):
    self._report, self._data_id = report, id(records)
    if "data" not in self._report._props:
      self._report._props["data"] = {"sources": {}, "schema": {}}
    self._report._props["data"]["sources"][self._data_id] = records
    self._report._props["data"]["schema"][self._data_id] = {'fncs': [], "profile": profile}
    self._data = self._report._props["data"]["sources"][self._data_id]
    self._schema = self._report._props["data"]["schema"][self._data_id]

  @classmethod
  def get(cls, report, varName):
    """
    Return the internal RawData object

    :return: A internal data object
    """
    return RawData(report, None)

  def setId(self, jqId):
    """
    Change the Id variable name for the javascript data source.

    Example

    :param jqId:

    :return: The Python object
    """
    self.jqId = jqId if jqId is not None else self._jqId
    return self

  def attach(self, html_obj, profile=False):
    """
    Attach the data object to a HTML Object.

    This function is automatically used in the different components in order
    to guarantee the link of the data. This will also ensure that the same data set will be store only once in the page

    Example

    :param html_obj:
    :param profile:

    """
    self._data["schema"][self._data_id].setdefault('containers', {})[html_obj.htmlId] = {'fncs': [], 'outs': None, "profile": profile}
    return ContainerData(self._report, self._data["schema"][self._data_id]['containers'][html_obj.htmlId])

  def toTsv(self, colNames=None, profile=False):
    """

    :return: A String with the Javascript function to be used
    """
    colNames = colNames or list(self._schema['keys'] | self._schema['values'])
    jsFncs = self._report._props.setdefault('js', {}).setdefault('functions', {})
    if "ToTsv" in jsFncs:
      jsFncs["ToTsv"] = {'content': '''var result = []; var tmp = []; var row = [];
                colNames.forEach(function(col){row.push(col)}); tmp.push(row.join('\t'));
                data.forEach(function(rec){
                  row = []; colNames.forEach(function(col){row.push(rec[col])});
                  tmp.push(row.join('\t'))}); result = tmp.join('\\n'); return result''', 'pmt': colNames}
    return "ToTsv(%s, %s)" % (self.jqId, colNames)

  @property
  def fnc(self):
    """

    :return:
    """
    return JsFncs.FncOnRecords(self, self._report._props, self._schema)

  @property
  def filter(self):
    """

    """
    return JsFncs.FncFiltere(self, self._report._props, self._schema)

  @property
  def to(self):
    """

    """
    return JsFncs.FncToObject(self, self._report._props, self._schema)

  @property
  def agg(self):
    """
    Property to the data aggregator functions
    The aggregator will create a new record with different column names
    """
    return JsFncs.FncRoAggRec(self, self._report._props, self._schema)

  def toStr(self):
    data = "data_%s" % self._data_id
    # Add the different javascript transformation functions
    for fnc in self._schema.get('fncs', []):
      data = fnc % data
    # Add the global filtering rules
    if len(self._schema.get("filters", [])) > 0:
      data = "%s(%s, [%s])" % ("JsFilter", data, ",".join(self._schema['filters']))
    # Add the final output object transformation
    if self._schema.get('out', None) is not None:
      data = self._schema['out'] % data
    return data


class JsData(object):

  def __init__(self, src):
    self._src = src

  def loop(self):
    return DataLoop()

  def reduce(self):
    return DataReduce()

  def sort(self):
    return DataSort()

  def each(self):
    return DataEach()

  @property
  def all(self):
    return DataAll()

  def crossfilter(self, data=None, var_name=None, crossDimension=None):
    """
    Description:
    -----------
    A crossfilter represents a multi-dimensional dataset.

    Constructs a new crossfilter. If records is specified, simultaneously adds the specified records.
    Records can be any array of JavaScript objects or primitives.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param data:
    """
    from epyk.core.js.packages.JsCrossFilter import CrossFilter

    if data is None:
      data = "%s.top(Infinity)" % crossDimension.toStr()

    if var_name is None:
      return CrossFilter(self._src, varName=var_name, data=data, setVar=False)

    return CrossFilter(self._src, varName=JsUtils.getJsValid(var_name), data=data)

  def dataset(self, data, var_name=None, options=None):
    """
    Description:
    -----------
    One of the starting points of the visualizations of vis.js is that they can deal with dynamic data, and allow manipulation of the data.
    To enable this, vis.js includes a flexible key/value based DataSet and DataView to handle unstructured JSON data.

    Related Pages:

			https://visjs.github.io/vis-data/data/index.html

    Attributes:
    ----------
    :param data:
    :param var_name:
    :param options:
    """
    vis_obj = VisDataSet(self._src, data=data, varName=JsUtils.getJsValid(var_name))
    if options is not None:
      vis_obj.setOptions(options)
    return vis_obj

  def dataview(self, dataset, var_name=None, options=None):
    """
    Description:
    -----------
    A DataView offers a filtered and/or formatted view on a DataSet.
    One can subscribe to changes in a DataView, and easily get filtered or formatted data without having to specify filters and field types all the time.

    Viz.Js module

    Related Pages:

			https://visjs.github.io/vis-data/data/dataview.html

    Attributes:
    ----------
    :param dataset:
    :param options:
    :param var_name:
    """
    vis_obj = VisDataView(self._src, data=dataset.varId, varName=JsUtils.getJsValid(var_name))
    if options is not None:
      vis_obj.setOptions(options)
    return vis_obj

  def records(self, data):
    """

    :param data:

    :return:
    """
    return RawData(self._src, data)

  @property
  def null(self):
    """
    Javascript null reference
    """
    return JsObject.JsObject("null", isPyData=False)


class JsDataTransfer(object):

  def __init__(self, varName):
    self.varId = varName

  @property
  def text(self):
    """
    Example

    Documentation

    :return:
    """
    return JsString.JsString("%s.getData('text')" % self.varId, isPyData=False)

  @property
  def files(self):
    """
    The DataTransfer.files property is a list of the files in the drag operation. If the operation includes no files, the list is empty.

    This feature can be used to drag files from a user's desktop to the browser.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/DataTransfer/files

    :return:
    """
    return JsArray.JsArray.get("%s.files" % self.varId)

  @property
  def dropEffect(self, flag=False):
    """

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/DataTransfer/dropEffect

    :return:
    """
    if flag == False:
      return JsBoolean.JsBoolean("%s.dropEffect" % self.varId)

    if flag not in [None, 'move', 'link', 'copy']:
      raise Exception("")

    flag = JsUtils.jsConvertData(flag, None)
    return JsFncs.JsFunction("%s.dropEffect = %s" % (self.varId, flag))

  @property
  def effectAllowed(self, flag=False):
    """

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/DataTransfer/effectAllowed

    :return:
    """
    if flag == False:
      return JsBoolean.JsBoolean("%s.effectAllowed" % self.varId)

    if flag not in [None, 'move', 'link', 'copy']:
      raise Exception("")

    flag = JsUtils.jsConvertData(flag, None)
    return JsFncs.JsFunction("%s.effectAllowed = %s" % (self.varId, flag))

  def clearData(self, format=None):
    """
    Remove the data associated with a given type. The type argument is optional.
    If the type is empty or not specified, the data associated with all types is removed.
    If data for the specified type does not exist, or the data transfer contains no data, this method will have no effect.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/DataTransfer/clearData

    :return:
    """
    if format is None:
      return JsFncs.JsFunction("%s.clearData()" % self.varId)

    format = JsUtils.jsConvertData(format, None)
    return JsFncs.JsFunction("%s.clearData(%s)" % (self.varId, format))

  def setDragImage(self):
    """

    :return:
    """

  def setData(self, data, format='text'):
    """

    :param data:
    :param format:
    :return:
    """
    format = JsUtils.jsConvertData(format, None)
    data = JsUtils.jsConvertData(data, None)
    return JsFncs.JsFunction("%s.setData(%s, %s)" % (self.varId, format, data))

  def getData(self, format="text"):
    """

    Example

    Documentation

    :param format:
    :return:
    """
    return JsString.JsString("%s.getData(%s)" % (self.varId, format), isPyData=False)



"""
Create a HTML page only from JavaScript functions
"""

from epyk.core.js import Js
from epyk.core.js import JsUtils
from epyk.core.js.packages import JsTabulator
from epyk.tests import test_statics

# Create a blank Javascript Python object
jsObj = Js.JsBase()

# Write the results to a Javascript file with a Launcher
f = JsUtils.JsFile("TestTabulator", path=test_statics.OUTPUT_PATHS)

tab = JsTabulator.Tabulator(selector="table", setVar=False)
tab.setData([])

col = JsTabulator.ColumnComponent()
col.delete()

f.writeJs([
  tab.getColumns.forEach(jsObj.console.log("column")).table.addRow({"test": "ok"}),
  col
])

# Close the file and print the location of the launcher
print(f.close(jsObj))


import os
import json
import importlib
import collections
import datetime
import time
import re
import logging

try:
  basestring
except NameError:
  basestring = str

from epyk.core.js import Imports
from epyk.interfaces import Components
from epyk.core.css.themes import Theme
from epyk.core.css import Classes

from epyk.core import html
from epyk.core import js
from epyk.core import py
from epyk.core import data

from epyk.core.html import symboles
from epyk.core.html import entities
from epyk.core.py import OrderedSet


class ContextRun(object):
  __slots__ = ['mac_address', 'host_name', 'current_user', 'report_name', 'script_name', 'local_path', 'url_root', 'title', 'is_local', 'url', 'username']
  __internals_attrs = ['mac_address', 'local_path', 'is_local', 'url', 'username']

  def __init__(self, report_name=None, script_name=None, current_user="local", host_name="script",
               url_root='http://127.0.0.1', title=None, urlMaps=None):
    self.report_name, self.script_name = report_name, script_name
    self.current_user, self.host_name, self.mac_address, self.url_root = current_user, host_name, "", None
    self.url = "#"
    if report_name is not None:
      mod = importlib.import_module('%s.__init__' % report_name)
      self.local_path, _ = os.path.split(os.path.abspath(mod.__file__))
    else:
      self.local_path = os.getcwd()
    self.title = self.script_name if title is None else title
    self.is_local = True

  def to_dict(self, inputs_scope=True):
    """

    :param inputs_scope: Boolean
    :return:
    """
    if inputs_scope:
      return dict([(s, getattr(self, s)) for s in self.__slots__ if s not in self.__internals_attrs])

    return dict([(s, getattr(self, s)) for s in self.__slots__])

  def __str__(self):
    return str(self.to_dict())


class Report(object):
  showNavMenu, withContainer = False, False
  ext_packages = None # For extension modules

  def __init__(self, run_options=None):
    #
    self._css, self._ui, self._js, self._py, self._theme, self.__body = {}, None, None, None, None, None
    self._props, self._tags, self._header_obj, self.__import_manage = {'js': {'onReady': OrderedSet()}}, None, None, None
    self.run = self.run_context(run_options if run_options is not None else {})

    self.timestamp, self.runTime = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S'), time.time() * 100
    self.content, self.shortcuts, self.exportCsv, self.jsSources = [], {}, {}, {}
    self._dbBindings, self._dbErrors = {}, collections.defaultdict(int)
    self.currentTitleObj, self.navBarContent, self.sideBarActions = {}, {'content': []}, []
    self.htmlItems, self.jsOnLoad, self.http, self.htmlCodes, self.htmlRefs = {}, [], {}, {}, {}
    self.notifications = collections.defaultdict(list)
    self.interruptReport, self._propagate = (False, None), []

    self.sourceDef, self.localFiles, self.libDef, self._run, self._scroll, self._contextMenu = {}, {}, {}, {}, set(), {}
    self.logo, self._dbSettings, self.dbsDef, self._cssText, self._jsText = None, None, {}, [], []

    #
    self.jsOnLoadFnc, self.jsWindowLoadFnc = OrderedSet(), OrderedSet()
    self.jsOnLoadEvtsFnc = OrderedSet()

    self.jsImports, self.cssImport = set(), set()
    self.jsLocalImports, self.cssLocalImports = set(), set()

  def run_context(self, run_options):
    """

    :return:
    """
    return ContextRun(**run_options)

  @property
  def body(self):
    """
    Property that returns the Body element of the HTML page
    """
    if self.__body is None:
      self.__body = html.Html.Body(self, None)
    return self.__body

  @body.setter
  def body(self, calc):
    self.__body = calc(self, None)

  @property
  def theme(self):
    """
    Return the currently used :doc:`report/theme` for the report
    """
    if self._theme is None:
      self._theme = Theme.ThemeDefault()
    return self._theme

  @theme.setter
  def theme(self, theme):
    if isinstance(theme, dict):
      self._theme = Theme.ThemeCustome()
      self._theme = theme
    else:
      self._theme = theme

  def imports(self, online=False):
    """
    Return the :doc:`report/import_manager`, which allows to import automatically packages for certain components to run.
    """
    if self.__import_manage is None:
      self.__import_manage = Imports.ImportManager(online, report=self)
    return self.__import_manage

  @property
  def symbols(self):
    """
    Description:
    ------------
    Shortcut to the HTML symbols

    Related Pages:

      https://www.w3schools.com/html/html_symbols.asp
      https://www.w3schools.com/charsets/ref_utf_math.asp

    Those can be added in string in order to improve the render of a text.
    """
    return symboles.Symboles()

  @property
  def entities(self):
    """
    Description:
    ------------
    Shortcut to the HTML Entities

    Related Pages:

      https://www.w3schools.com/html/html_entities.asp

    Those can be added in string in order to improve the render of a text.
    """
    return entities.Entities()

  @property
  def ui(self):
    """
    Description:
    ------------
    User Interface section.

    All the :doc:`components <report/ui>` which can be used in the dashboard to display the data.
    Within this object different categories of items can be used like (list, simple text, charts...)

    Related Pages:

	    https://www.w3schools.com/html/default.asp

    :rtype: :doc:`Components.Components <report/ui>`

    :return: Python HTML object
    """
    if self._ui is None:
      self._ui = Components.Components(self)
    return self._ui

  @property
  def css(self):
    """
    Returns the set of :doc:`CSS Classes <css>` for the HTML report
    """
    return Classes.Catalog(self, {'other': set()})._class_type('other')

  @property
  def js(self):
    """
    Description:
    ------------
    Go to the Javascript section. Property to get all the JavaScript features.
    Most of the standard modules will be available in order to add event and interaction to the Js transpiled

    Usage::

      js.console.log("test")

    Related Pages:

      https://www.w3schools.com/js/default.asp

    :return: Python HTML object
    """
    if self._js is None:
      self._js = js.Js.JsBase(self)
    return self._js

  @property
  def py(self):
    """
    Description:
    ------------
    Python external module section.

    Related Pages:

      https://www.w3schools.com/js/default.asp

    :return: Python HTML object
    """
    if self._py is None:
      self._py = py.PyExt.PyExt(self)
    return self._py

  @property
  def data(self):
    """
    Python internal data source management

    This can be extended by inheriting from this epyk.core.data.DataSrc.DataSrc
    and adding extra entry points

    :return: The framework available data source
    """
    return data.Data.DataSrc(self)

  def itemFromCode(self, htmlCode):
    """

    :param htmlCode:

    :rtype: html.Html.Html
    :return:
    """
    return self.htmlCodes[htmlCode]

  def item(self, itemId):
    """

    :param itemId:
    :return:
    """
    return self.htmlItems[itemId]

  def socketSend(self, htmlCode, data, report_name=None, script_name=None):
    try:
      from urllib.parse import urlparse, urlencode
      from urllib.request import urlopen, Request
      from urllib.error import HTTPError
    except ImportError:
      from urlparse import urlparse
      from urllib import urlencode
      from urllib2 import urlopen, Request, HTTPError

    urls = [htmlCode]
    if report_name or report_name is None:
      urls.append(self.run.report_name if report_name is None else report_name)
      if script_name or script_name is None:
        urls.append(self.run.script_name if script_name is None else script_name)
    response = urlopen(Request("%s%smessage/%s" % (self.run.url_root, self._urlsApp["index"].replace("/index", "/"), "/".join(urls)),
                    data=urlencode({'data': json.dumps(data)}).encode('utf-8')))
    response.read()

  # -----------------------------------------------------------------------------------------
  #                                    LOGGING AND DEBUGGING FUNCTIONS
  # -----------------------------------------------------------------------------------------
  def log(self, text, type='DEBUG'):
    """
    This will write some logs to your local user.log file in \system\logs.
    """
    log = logging.getLogger('user')
    if type.upper() == 'INFO':
      log.info("| %s [ %s ] >> %s" % (self.run.report_name, self.run.script_name, text))
    elif type.upper() == 'WARNING':
      log.warning("| %s [ %s ] >> %s" % (self.run.report_name, self.run.script_name, text))
    else:
      log.debug("| %s [ %s ] >> %s" % (self.run.report_name, self.run.script_name, text))

  @property
  def outs(self):
    """

    :return:
    """
    return py.PyOuts.PyOuts(self)

  @property
  def headers(self):
    """
    Property to the HTML page header
    """
    if self._header_obj is None:
      self._header_obj = html.Header.Header(self)
    return self._header_obj

  def to_html(self, online=False, mode=None, html_template=None, fnc=None):
    """

    :param online:
    :param mode:
    :param html_template:
    :param fnc:

    :return:
    """
    if html_template is None:
      html_template = html.templates.HtmlTmplBase.DATA

    if fnc is not None:
      return fnc(html_template, **self.html(online, mode))

    return html_template % self.html(online, mode)

  def to_html_error(self, tracebackStr, server_owner, title="Script Error Details", icon="fas fa-exclamation-triangle",
                    pyModule=None, online=False, mode=None, html_template=None, fnc=None):
    """

    :param tracebackStr:
    :param server_owner:
    :param title:
    :param icon:
    :param pyModule:

    :return:
    """
    report_error = Report(self.run.to_dict())
    report_error._urlsApp.update(self._urlsApp)
    report_error.nav(selected='Reports', breadcrum=True)
    for notification in self._propagate:
      report_error.add(notification)
    report_error.ui.text("<i class='fas fa-cog'></i>&nbsp;To open the report again click [here](/reports/run/%s/%s)" % (
      report_error.run.report_name, report_error.run.script_name)).css('margin', '5px')
    owners = ";".join(pyModule.OWNERS) if pyModule is not None and hasattr(pyModule, 'OWNERS') else server_owner
    report_error.ui.text("<i class='fas fa-envelope'></i>&nbsp;To raise this issue to the report owner <a onclick='window.NO_UNLOAD = true;' href='mailto:%(owners)s?subject=Report error&body=%(error)s'>contact us</a>" % {
      'owners': owners, 'error': tracebackStr.replace("'", "").replace('"', "")}).css('margin', '5px')
    report_error.ui.rich.textborder({'title': title, 'icon': icon, 'value': "<pre style='padding:5px;white-space:pre-wrap; '>%s</pre>" % '<br>'.join([re.sub(r'^ +', lambda m: ' ' * len(m.group()), i) for i in tracebackStr.split('\n')]),
                                     'color': self.getColor('danger', 1), 'colorTitle': self.getColor('danger', 1)})
    return report_error.to_html(online, mode, html_template, fnc)

  def dumps(self, data):
    """
    Function used to dump the data before being sent to the Javascript layer
    This function relies on json.dumps with a special encoder in order to work with Numpy array and Pandas data structures.

    As NaN is not valid on the Json side those object are not allowed during the dump.
    It is advised to use fillna() in your script before returning the data to the framework to avoid this issue.

    Example::

      report.dumps(result)

    Related Pages:

			https://docs.python.org/2/library/json.html

    :param data: The python dictionary or data structure
    :return: The serialised data
    """
    return json.dumps(data, cls=js.JsEncoder.Encoder, allow_nan=False)



"""
CSS Style module for the Label components
"""

from epyk.core.css.styles.classes import CssStyle


# class CssLabelContainer(CssStyle.Style):
#   _attrs = {'display': 'block', 'position': 'relative', 'cursor': 'pointer', '-webkit-user-select': 'none',
#             '-moz-user-select': 'none', '-ms-user-select': 'none', 'user-select': 'none'}
#
#   def customize(self):
#     self.css({'font-family': Defaults_css.Font.family,
#               'font-size': '%s%s' % (Defaults_css.Font.size, Defaults_css.Font.unit)})
#

class CssLabelContainerDisabled(CssStyle.Style):
  _attrs = {'color': 'red', 'cursor': 'not-allowed'}

  def customize(self):
    self.css({'color': self.rptObj.theme.danger[1]})


class CssLabelCheckMarkHover(CssStyle.Style):
  _selectors = {'child': "label"}

  def customize(self):
    self.hover.css({'background-color': self.rptObj.theme.colors[5]})




import sys

from epyk.core.data import DataClass


class Options(DataClass):

  def __init__(self, report, attrs=None, options=None):
    super(Options, self).__init__(report, attrs, options)
    if attrs is not None:
      for k, v in attrs.items():
        if hasattr(self, k):
          setattr(self, k, v)
        else:
          self._report._jsStyles[k] = v

  def _config_get(self, dflt=None, name=None):
    """
    Description:
    ------------
    Get the option attribute to be added on the Javascript side during the component build

    Attributes:
    ----------
    :param name: String. The attribute name
    """
    return self._report._jsStyles.get(name or sys._getframe().f_back.f_code.co_name, dflt)

  def _config(self, value, name=None):
    """
    Description:
    ------------
    Set the option attribute to be added on the Javascript side during the component build

    Attributes:
    ----------
    :param value: Object. The value for the name
    :param name: String. The attribute name
    """
    self._report._jsStyles[name or sys._getframe().f_back.f_code.co_name] = value





import sys
import importlib

from epyk.core.py import PyCrypto
from epyk.core.py import PyHash
from epyk.core.py import PyRest
from epyk.core.py import PyDates
from epyk.core.py import PyMarkdown

from epyk.core.js.Imports import requires

#
# http://www.web-source.net/symbols.htm#.XiyBZuBzw2w
ENTITY_MAP = {"é": "&#233;", "È": "&#200;", "à": "&agrave;"}


class PyExt(object):
  class __internal(object):
    _props = {}

  def __init__(self, report=None):
    self._report = report if report is not None else self.__internal()
    self.hash = PyHash.SipHash().hashId
    self.today = self.dates.today
    self.now = self.dates.now
    self.cob = self.dates.cob
    self.request = self.requests.request
    if not 'py' in self._report._props:
      self._report._props['py'] = {}

  @property
  def dates(self):
    """
    This is a simple wrapper to the datetime Python module.

    No external package is required to use this interface.
    :return: A PyDate object
    """
    return PyDates.PyDates(self._report)

  @property
  def requests(self):
    """
    This is a simple wrapper to the internal Python modules to run REST calls.

    No external package is required to use this interface.
    :return: A PyRest object
    """
    return PyRest.PyRest(self._report)

  @property
  def crypto(self):
    """
    Property to the internal cryptography module.

    This will rely on the package cryptography. This should be added to the python environment before using it.
    This package can be installed using the usual pip install function.

    Related Pages:

			https://pypi.org/project/cryptography/
    https://cryptography.io/en/latest/

    :return:
    """
    return PyCrypto.PyCrypto(self._report)

  @property
  def markdown(self):
    """ Property to the Markdown String conversion """
    return PyMarkdown.Convertor(self._report)

  def import_lib(self, lib_name, folder="libs", report_name=None, path=None):
    """
    Import dynamically a python module

    Example
    rptObj.py.import_lib("test.py", folder="tables", path=r"filePath")

    :param lib_name: The python module name
    :param folder: The internal folder with the libraries to be imported
    :param report_name: Optional, the report name in which the library is defined. Default current folder
    :param path: Optional, the path to be added to the classpath

    :return: The imported Python module
    """
    lib_name = lib_name.replace(".py", "")
    if path is not None:
      if path not in sys.path:
        sys.path.append(path)
      return importlib.import_module('%s.%s' % (folder, lib_name))

    else:
      if report_name is None:
        return importlib.import_module('%s.%s.%s' % (self._report.run.report_name, folder, lib_name))

      return importlib.import_module('%s.%s.%s' % (report_name, folder, lib_name))

  def import_package(self, package, sub_module=None):
    """
    Install the external Python package.
    This can automatically installed it from the Python Index online repository is missing

    Example
    >>> PyExt().import_package("sqlalchemy").__name__
    'sqlalchemy'

    :param package: The Python Package Name
    :param sub_module: The sub module or class within the package

    :return: The installed Python module
    """
    package_alias = "%s_%s" % (package, sub_module) if sub_module is not None else package
    if not package_alias in self._report._props:
      self._report._props[package_alias] = requires(package, reason='Missing Package', package=sub_module, install=package)
    return self._report._props[package_alias]

  def clean(self, text):
    """
    Replace the special characters by the corresponding HTML entities

    :param text:
    :return:
    """
    for c, t in ENTITY_MAP.items():
      if c in text:
        text = text.replace(c, t)
    return text





from epyk.core.css.catalogs import Catalog

from epyk.core.css.styles.classes import CssStylesButton


class CatalogButton(Catalog.CatalogGroup):

  def basic(self):
    """
    Description:
    -----------
    Basic style for a button
    """
    return self._set_class(CssStylesButton.CssButtonBasic)

  def important(self):
    """
    Description:
    -----------
    Important style for a button
    """
    return self._set_class(CssStylesButton.CssButtonImportant)

  def border_rounded(self):
    """
    Description:
    -----------

    """
    pass

  def reset(self):
    """
    Description:
    -----------
    """
    return self._set_class(CssStylesButton.CssButtonReset)

  def success(self):
    """
    Description:
    -----------
    """
    return self._set_class(CssStylesButton.CssButtonSuccess)

  def content(self):
    """
    Description:
    -----------
    CSS Class for the underlying item panel.
    By default this item is not visible and this will change when the mouse is hover
    """
    return self._set_class(CssStylesButton.CssButtonContentHover)

  def content_link(self):
    """
    Description:
    -----------
    CSS Class for the item link.
    This will set the background color when the mouse is hover.

    The color is deduced from the defined theme
    """
    return self._set_class(CssStylesButton.CssButtonContentAHover)




from epyk.core.css.styles.classes import CssStyle


class CssBorderRadius(CssStyle.Style):
  _attrs = {'border-radius': '5px'}



class CssButtonBasic(CssStyle.Style):
  # Static properties for this class
  _attrs = {'font-weight': 'bold', 'padding': '2px 20px', 'margin': '2px 0 2px 0', 'text-decoration': 'none',
            'border-radius': '4px', 'white-space': 'nowrap', 'display': 'inline-block', 'line-height': '30px',
            '-webkit-appearance': 'none', '-moz-appearance': 'none'}
  _hover = {'text-decoration': 'none', 'cursor': 'pointer'}
  _focus = {'outline': 0}
  _disabled = {'cursor': 'none'}

  def customize(self):
    self.css({'border': '1px solid %s' % self.rptObj.theme.greys[4], 'color': self.rptObj.theme.colors[-1],
              'background-color': self.rptObj.theme.greys[0]})
    self.hover.css({'background-color': self.rptObj.theme.colors[0], 'color': self.rptObj.theme.colors[-1]})
    self.disabled.css({'background-color': self.rptObj.theme.colors[-1], 'color': self.rptObj.theme.colors[6],
                       'font-style': 'italic'})


class CssButtonImportant(CssStyle.Style):
  # Static properties for this class
  _attrs = {'font-weight': 'bold', 'padding': '2px 20px', 'margin': '2px 0 2px 0', 'text-decoration': 'none',
            'border-radius': '4px', 'white-space': 'nowrap', 'display': 'inline-block', 'line-height': '30px',
            '-webkit-appearance': 'none', '-moz-appearance': 'none'}
  _hover = {'text-decoration': 'none', 'cursor': 'pointer'}
  _focus = {'outline': 0}
  _disabled = {'cursor': 'none'}

  def customize(self):
    self.css({'border': '1px solid', 'color': self.rptObj.theme.greys[0],
              'background-color': self.rptObj.theme.colors[-2]})
    self.hover.css({'background-color': self.rptObj.theme.colors[-1], 'color': self.rptObj.theme.greys[0]})
    self.disabled.css({'background-color': self.rptObj.theme.colors[-1], 'color': self.rptObj.theme.colors[6],
                       'font-style': 'italic'})


class CssButtonReset(CssStyle.Style):
  # Static properties for this class
  _attrs = {'font-weight': 'bold', 'padding': '5px 10px 5px 10px', 'margin-top': '5px', 'text-decoration': 'none',
            'border-radius': '5px', 'display': 'inline-block', 'text-transform': 'uppercase'}
  _hover = {'text-decoration': 'none', 'cursor': 'pointer'}

  def customize(self):
    self.css({'border': '1px solid %s' % self.rptObj.theme.danger[1], 'color': self.rptObj.theme.danger[1],
              'background-color': self.rptObj.theme.greys[0]})
    self.hover.css({'background-color': self.rptObj.theme.danger[1], 'color': self.rptObj.theme.greys[0]})


class CssButtonSuccess(CssStyle.Style):
  # Static properties for this class
  _attrs = {'font-weight': 'bold', 'padding': '10px 10px 10px 10px', 'margin': '10px 0px 10px 5px',
            'text-decoration': 'none', 'border-radius': '5px', 'display': 'inline-block', 'text-transform': 'uppercase'}
  _hover = {'text-decoration': 'none', 'cursor': 'pointer'}

  def customize(self):
    self.css({'color': self.rptObj.theme.colors[9], 'background-color': self.rptObj.theme.greys[0],
              'border': '1px solid %s' % self.rptObj.theme.colors[9]})
    self.hover.css({'color': self.rptObj.theme.greys[0], 'background-color': self.rptObj.theme.colors[9]})


class CssButtonContentHover(CssStyle.Style):
  _hover = {'display': 'block !IMPORTANT'}
  _selectors = {'suffix': ".dropdown-content"}


class CssButtonContentAHover(CssStyle.Style):
  _selectors = {'child': ".dropdown-content a"}

  def customize(self):
    self.hover.css({'background-color': self.rptObj.theme.success[0]})



"""

"""

from epyk.core.css.catalogs import Catalog

from epyk.core.css.styles.classes import CssStylesPopup


class CatalogPopup(Catalog.CatalogGroup):
  def table(self):
    """  """
    return self._set_class(CssStylesPopup.CssPopupTable)

  def table_title(self):
    """  """
    return self._set_class(CssStylesPopup.CssPopupTableTitle)

  def table_content(self):
    """  """
    return self._set_class(CssStylesPopup.CssPopupTableTitleContent)

  def loading(self):
    """  """
    return self._set_class(CssStylesPopup.CssEventLoading)



"""
Related Pages:

		https://www.w3schools.com/charsets/ref_utf_letterlike.asp
"""

ACCOUNT_OF = '&#8448'
ADDRESSED_TO_THE_SUBJECT = '&#8449'
DOUBLE_STRUCK_CAPITAL_C = '&#8450'
DEGREE_CELSIUS = '&#8451'
CENTRE_LINE_SYMBOL = '&#8452'
CARE_OF = '&#8453'
CADA_UNA = '&#8454'
EULER_CONSTANT = '&#8455'
SCRUPLE = '&#8456'
DEGREE_FAHRENHEIT = '&#8457'
SCRIPT_SMALL_G = '&#8458'
SCRIPT_CAPITAL_H = '&#8459'
BLACK_LETTER_CAPITAL_H = '&#8460'
DOUBLE_STRUCK_CAPITAL_H = '&#8461'
PLANCK_CONSTANT = '&#8462'
PLANCK_CONSTANT_OVER_TWO_PI = '&#8463'
SCRIPT_CAPITAL_I = '&#8464'
BLACK_LETTER_CAPITAL_I = '&#8465'
SCRIPT_CAPITAL_L = '&#8466'
SCRIPT_SMALL_L = '&#8467'
L_B_BAR_SYMBOL = '&#8468'
DOUBLE_STRUCK_CAPITAL_N = '&#8469'
NUMERO_SIGN = '&#8470'
SOUND_RECORDING_COPYRIGHT = '&#8471'
SCRIPT_CAPITAL_P = '&#8472'
DOUBLE_STRUCK_CAPITAL_P = '&#8473'
DOUBLE_STRUCK_CAPITAL_Q = '&#8474'
SCRIPT_CAPITAL_R = '&#8475'
BLACK_LETTER_CAPITAL_R = '&#8476'
DOUBLE_STRUCK_CAPITAL_R = '&#8477'
PRESCRIPTION_TAKE = '&#8478'
RESPONSE = '&#8479'
SERVICE_MARK = '&#8480'
TELEPHONE_SIGN = '&#8481'
TRADE_MARK_SIGN = '&#8482'
VERSICLE = '&#8483'
DOUBLE_STRUCK_CAPITAL_Z = '&#8484'
OUNCE_SIGN = '&#8485'
OHM_SIGN = '&#8486'
INVERTED_OHM_SIGN = '&#8487'
BLACK_LETTER_CAPITAL_Z = '&#8488'
TURNED_GREEK_SMALL_LETTER_IOTA = '&#8489'
KELVIN_SIGN = '&#8490'
ANGSTROM_SIGN = '&#8491'
SCRIPT_CAPITAL_B = '&#8492'
BLACK_LETTER_CAPITAL_C = '&#8493'
ESTIMATED_SYMBOL = '&#8494'
SCRIPT_SMALL_E = '&#8495'
SCRIPT_CAPITAL_E = '&#8496'
SCRIPT_CAPITAL_F = '&#8497'
TURNED_CAPITAL_F = '&#8498'
SCRIPT_CAPITAL_M = '&#8499'
SCRIPT_SMALL_O = '&#8500'
ALEF_SYMBOL = '&#8501'
BET_SYMBOL = '&#8502'
GIMEL_SYMBOL = '&#8503'
DALET_SYMBOL = '&#8504'
INFORMATION_SOURCE = '&#8505'
ROTATED_CAPITAL_Q = '&#8506'
FACSIMILE_SIGN = '&#8507'
DOUBLE_STRUCK_SMALL_PI = '&#8508'
DOUBLE_STRUCK_SMALL_GAMMA = '&#8509'
DOUBLE_STRUCK_CAPITAL_GAMMA = '&#8510'
DOUBLE_STRUCK_CAPITAL_PI = '&#8511'
DOUBLE_STRUCK_N_ARY_SUMMATION = '&#8512'
TURNED_SANS_SERIF_CAPITAL_G = '&#8513'
TURNED_SANS_SERIF_CAPITAL_L = '&#8514'
REVERSED_SANS_SERIF_CAPITAL_L = '&#8515'
TURNED_SANS_SERIF_CAPITAL_Y = '&#8516'
DOUBLE_STRUCK_ITALIC_CAPITAL_D = '&#8517'
DOUBLE_STRUCK_ITALIC_SMALL_D = '&#8518'
DOUBLE_STRUCK_ITALIC_SMALL_E = '&#8519'
DOUBLE_STRUCK_ITALIC_SMALL_I = '&#8520'
DOUBLE_STRUCK_ITALIC_SMALL_J = '&#8521'
PROPERTY_LINE = '&#8522'
TURNED_AMPERSAND = '&#8523'
PER_SIGN = '&#8524'
AKTIESELSKAB = '&#8525'
TURNED_SMALL_F = '&#8526'
SYMBOL_FOR_SAMARITAN_SOURCE = '&#8527'



"""

"""

from epyk.core.js import Js
from epyk.core.js import JsUtils
from epyk.tests import test_statics

# Create a blank Javascript Python object
jsObj = Js.JsBase()

# Write the results to a Javascript file with a Launcher
f = JsUtils.JsFile("DataExamplesPlolty", path=test_statics.OUTPUT_PATHS)


# Write the Javascript fragments to the file
f.writeJs([

])

import json

#data = test_statics.get_data("stocks.json")

# Close the file and print the location of the launcher
print(f.close(jsObj))





from epyk.core.js.packages import packageImport
from epyk.core.data import DataGroup


class ExtsValidators(DataGroup):

  def custom(self, validator, moduleAlias):
    """
    Description:
    -----------

    http://tabulator.info/docs/4.0/mutators

    :param validator:
    :param moduleAlias:
    """
    self._report.jsImports.add(moduleAlias)
    self._attrs["validator"] = validator
    return self



"""
DC.js API

http://dc-js.github.io/dc.js/docs/html/
"""

from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObject
from epyk.core.js.packages import JsPackage


class DC(JsPackage):
  lib_alias = {'css': 'dc', 'js': 'dc'}

  def __init__(self, src=None, varName=None, setVar=True, parent=None):
    self.src, self._sub_chart = src, None
    if parent is not None:
      # for series chart the selector is specific
      self._selector = "new dc.%s('#%s')" % (self.chartFnc, parent.htmlId)
    self.varName, self.setVar = varName, setVar
    self.src.jsImports.add(self.lib_alias['js'])
    self.src.cssImport.add(self.lib_alias['css'])
    self._js, self._xaxis, self._yaxis, self._u = [[]], None, None, {}

  def x(self):
    return self.fnc("x(d3.scaleLinear().domain([0,20]))")

  def chartGroup(self, group_name):
    return self.fnc("chartGroup(%s)" % JsUtils.jsConvertFncs(group_name, toStr=True))

  def redrawGroup(self):
    """
    Redraws all charts in the same group as this chart, typically in reaction to a filter change. If the chart has a commitHandler, it will be executed and waited for.

    http://dc-js.github.io/dc.js/docs/html/BaseMixin.html#redrawGroup__anchor
    """
    return JsObject.JsObject.get("%s.redrawGroup()" % self.varId)

  def width(self, n):
    """

    :param n:

    :return: Return 'self' to allow the cnains on the Python side
    """
    return self.fnc("width(%s)" % JsUtils.jsConvertData(n, None))

  def height(self, n):
    """

    :param n:
    :return:
    """
    return self.fnc("height(%s)" % JsUtils.jsConvertData(n, None))

  def yAxisLabel(self, text):
    """

    :param text:

    :return:
    """
    return self.fnc("yAxisLabel(%s)" % JsUtils.jsConvertData(text, None))

  def xAxisLabel(self, text):
    """

    :param text:
    :return:
    """
    return self.fnc("xAxisLabel(%s)" % JsUtils.jsConvertData(text, None))

  def render(self):
    """

    :return:
    """
    self._js.append(["render()"])
    return self

  def dimension(self, values):
    return self.fnc("dimension(%s)" % values)
    #return self.fnc("dimension(%s)" % JsUtils.jsConvertData(values, None))

  def group(self, groups):
    return self.fnc("group(%s)" % groups)
    #return self.fnc("group(%s)" % JsUtils.jsConvertData(groups, None))

  def toStr(self):
    """
    Description:
    ------------
    Javascript representation

    :return: Return the Javascript String
    """
    str_chart = ""
    if self._sub_chart is not None:
      if self._sub_chart._js != [[]]:
        str_chart = ".chart(function(c) { return %s; })" % self._sub_chart.toStr()
      else:
        str_chart = ".chart(function(c) { return %s; })" % self._sub_chart._selector

    if self._selector is None:
      raise Exception("Selector not defined, use this() or new() first")

    obj_content = []
    for i, js in enumerate(self._js):
      if len(js) == 0:
        continue

      str_fnc = ".".join([d.toStr() if hasattr(d, "toStr") else d for d in js])
      if self.setVar:
        if str_fnc:
          str_fnc = "var %s = %s; %s%s.%s" % (self.varId, self._selector, self.varId, str_chart, str_fnc)
        else:
          str_fnc = "var %s = %s" % (self.varId, self._selector)
        self.setVar = False
      else:
        if str_fnc:
          if i in self._u:
            # to avoid raising an error when the variable is not defined
            str_fnc = "if(%s !== undefined){%s.%s}" % (self.varId, self.varId, str_fnc)
          else:
            varId = self._mapVarId(str_fnc, self.varId)
            str_fnc = "%s.%s" % (varId, str_fnc)
        else:
          str_fnc = self.varId
      obj_content.append(str_fnc)
    self._js = [[]] # empty the stack
    return "; ".join(obj_content)


class Line(DC):
  chartFnc = "lineChart"

  def curve(self, fnc):
    return self.fnc("curve(%s)" % fnc)

  def curveStepBefore(self):
    return self.fnc("curve(d3.curveStepBefore)")

  def renderArea(self, flag):
    """

    :param flag:
    """
    return self.fnc("renderArea(%s)" % JsUtils.jsConvertData(flag, None))

  def renderDataPoints(self, flag):
    """

    :param flag:
    :return:
    """
    return self.fnc("renderDataPoints(%s)" % JsUtils.jsConvertData(flag, None))

  def clipPadding(self, value):
    """

    :param value:
    :return:
    """
    return self.fnc("clipPadding(%s)" % JsUtils.jsConvertData(value, None))


class Bar(DC):
  chartFnc = "barChart"

  def controlsUseVisibility(self, flag):
    """

    https://dc-js.github.io/dc.js/examples/bar-single-select.html

    :param flag:
    """
    return self.fnc("controlsUseVisibility(%s)" % JsUtils.jsConvertData(flag, None))

  def singleSelection(self):
    self.xUnitsOridinal()
    self.scaleBand()
    return self.fnc("addFilterHandler(function(filters, filter) {return [filter]; })")

  def xUnitsOridinal(self):
    return self.fnc("xUnits(dc.units.ordinal)")

  def scaleBand(self):
    return self.fnc("x(d3.scaleBand())")


class Row(DC):
  chartFnc = "rowChart"

  def elasticY(self, flag):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/row-targets.html

    :param flag:
    """
    return self.fnc("elasticY(%s)" % JsUtils.jsConvertData(flag, None))

  def elasticX(self, flag):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/row-targets.html

    :param flag:
    """
    return self.fnc("elasticX(%s)" % JsUtils.jsConvertData(flag, None))

  def yAxisLabel(self, text):
    raise Exception("")


class Pie(DC):
  chartFnc = "pieChart"

  def slicesCap(self, value):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/pie.html

    :param value:
    """
    return self.fnc("slicesCap(%s)" % JsUtils.jsConvertData(value, None))

  def innerRadius(self, value):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/pie.html

    :param value:
    """
    return self.fnc("innerRadius(%s)" % JsUtils.jsConvertData(value, None))

  def legend(self):
    pass

  def drawPaths(self, flag):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/pie-external-labels.html

    :param flag:
    """
    return self.fnc("drawPaths(%s)" % JsUtils.jsConvertData(flag, None))

  def externalLabels(self, size):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/pie-external-labels.html

    :param size:
    """
    return self.fnc("externalLabels(%s)" % JsUtils.jsConvertData(size, None))

  def externalRadiusPadding(self, size):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/pie-external-labels.html

    :param size:
    """
    return self.fnc("externalRadiusPadding(%s)" % JsUtils.jsConvertData(size, None))


class Series(DC):
  chartFnc = "seriesChart"

  def line(self):
    self._sub_chart = Line(src=self.src, setVar=False)
    self._sub_chart._selector = "new dc.lineChart(c)"
    return self

  def scatter(self):
    self._sub_chart = Scatter(src=self.src, setVar=False)
    self._sub_chart._selector = "new dc.scatterPlot(c)"
    return self

  def bubble(self):
    self._sub_chart = Bubble(src=self.src, setVar=False)
    self._sub_chart._selector = "new dc.bubbleChart(c)"
    return self

  def bar(self):
    self._sub_chart = Bar(src=self.src, setVar=False)
    self._sub_chart._selector = "new dc.barChart(c)"
    return self

  def seriesAccessor(self, jsFncs):
    """

    :param jsFncs:
    """
    return self.fnc("seriesAccessor(function(d) {%s; })" % JsUtils.jsConvertFncs(jsFncs, toStr=True))

  def seriesAccessorByKey(self, index=None, str_format=None):
    """

    :param index:
    :param str_format:
    """
    if str_format is not None:
      key = str_format % ("d.key[%s]" % index)
      return self.fnc("seriesAccessor(function(d) {return %s; })" % key)

    if index is None:
      return self.fnc("seriesAccessor(function(d) { return d.key; })")

    return self.fnc("seriesAccessor(function(d) { return d.key[%s]; })" % JsUtils.jsConvertData(index, None))

  def keyAccessor(self, index=None):
    if index is None:
      return self.fnc("keyAccessor(function(d) {return +d.key; })")

    return self.fnc("keyAccessor(function(d) {return +d.key[%s]; })" % index)

  def valueAccessor(self):
    return self.fnc("valueAccessor(function(d) {return d.value; })")

  def elasticY(self, flag):
    """

    :param flag:
    """
    return self.fnc("elasticY(%s)" % JsUtils.jsConvertData(flag, None))

  def mouseZoomable(self, flag):
    """

    :param flag:
    """
    return self.fnc("mouseZoomable(%s)" % JsUtils.jsConvertData(flag, None))

  def shareTitle(self, flag):
    """

    :param flag:
    """
    return self.fnc("shareTitle(%s)" % JsUtils.jsConvertData(flag, None))

  def controlsUseVisibility(self, flag):
    """

    https://dc-js.github.io/dc.js/examples/bar-single-select.html

    :param flag:
    """
    return self.fnc("controlsUseVisibility(%s)" % JsUtils.jsConvertData(flag, None))

  def singleSelection(self):
    self.xUnitsOridinal()
    self.scaleBand()
    return self.fnc("addFilterHandler(function(filters, filter) {return [filter]; })")

  def xUnitsOridinal(self):
    return self.fnc("xUnits(dc.units.ordinal)")

  def scaleBand(self):
    return self.fnc("x(d3.scaleBand())")

  def renderArea(self, flag):
    """

    :param flag:
    """
    return self.fnc("renderArea(%s)" % JsUtils.jsConvertData(flag, None))


class Scatter(DC):
  chartFnc = "scatterPlot"

  def radiusValueAccessorByKey(self, index=None, str_format=None):
    """

    :param index:
    :param str_format:
    """
    if str_format is not None:
      key = str_format % ("d.key[%s]" % index)
      return self.fnc("radiusValueAccessor(function(d) {return %s; })" % key)

    if index is None:
      return self.fnc("radiusValueAccessor(function(d) { return d.key; })")

    return self.fnc("radiusValueAccessor(function(d) { return d.key[%s]; })" % JsUtils.jsConvertData(index, None))

  def elasticX(self, flag):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/splom.html

    :param flag:
    """
    return self.fnc("elasticX(%s)" % JsUtils.jsConvertData(flag, None))

  def excludedOpacity(self, value):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/scatter-brushing.html

    :param value:
    """
    return self.fnc("excludedOpacity(%s)" % JsUtils.jsConvertData(value, None))

  def excludedColor(self, color):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/scatter-brushing.html

    :param color:
    """
    return self.fnc("excludedColor(%s)" % JsUtils.jsConvertData(color, None))

  def symbolSize(self, size):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/scatter-brushing.html

    :param color:
    """
    return self.fnc("symbolSize(%s)" % JsUtils.jsConvertData(size, None))

  def clipPadding(self, value):
    """

    :param value:
    :return:
    """
    return self.fnc("clipPadding(%s)" % JsUtils.jsConvertData(value, None))


class Bubble(Scatter):
  chartFnc = "bubbleChart"

  def keyAccessor(self, index=None):
    if index is None:
      return self.fnc("keyAccessor(function(d) {return +d.key; })")

    return self.fnc("keyAccessor(function(d) {return +d.key[%s]; })" % index)

  def radiusValueAccessorByKey(self, index=None, str_format=None, statc_factor=None):
    """

    :param index:
    :param str_format:
    """
    if str_format is not None:
      key = str_format % ("d.key[%s]" % index)
      return self.fnc("radiusValueAccessor(function(d) {return %s; })" % key)

    str_k = "d.key" if index is None else "d.key[%s]" % JsUtils.jsConvertData(index, None)
    if statc_factor is not None:
      str_k = "%s %s" % (str_k, statc_factor)

    return self.fnc("radiusValueAccessor(function(d) { return %s; })" % str_k)


class Sunburst(DC):
  chartFnc = "sunburstChart"

  def innerRadius(self, value):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/sunburst-equal-radii.html

    :param value:
    """
    return self.fnc("innerRadius(%s)" % JsUtils.jsConvertData(value, None))

  def ringSizes(self, jsFnc):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/sunburst-equal-radii.html

    :param jsFnc:
    """
    return self.fnc("ringSizes(%s)" % jsFnc)

  def equalRingSizes(self):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/sunburst-equal-radii.html

    :return:
    """
    return self.fnc("ringSizes(%s.equalRingSizes())" % self.varId)


class Composite(DC):
  chartFnc = "compositeChart"

  def x(self):
    pass

  def xUnits(self):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/pareto-chart.html
    """
    pass

  def elasticY(self, flag):
    """

    :param flag:
    """
    return self.fnc("elasticY(%s)" % JsUtils.jsConvertData(flag, None))

  def renderHorizontalGridLines(self, flag):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/composite-bar-line.html

    :param flag:
    """
    return self.fnc("renderHorizontalGridLines(%s)" % JsUtils.jsConvertData(flag, None))

  def compose(self, dc_charts):
    pass


class BoxPlot(DC):
  chartFnc = "boxPlot"

  def elasticY(self, flag):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/boxplot-basic.html

    :param flag:
    """
    return self.fnc("elasticY(%s)" % JsUtils.jsConvertData(flag, None))

  def elasticX(self, flag):
    """

    https://github.com/dc-js/dc.js/blob/master/web-src/examples/boxplot-basic.html

    :param flag:
    """
    return self.fnc("elasticX(%s)" % JsUtils.jsConvertData(flag, None))


class GeoChoropleth(DC):
  chartFnc = "geoChoroplethChart"



"""

"""

from epyk.core.py import PyRest


pyrest = PyRest.PyRest()
# content = pyrest.webscrapping(r"https://cdnjs.cloudflare.com/ajax/libs/jqueryui/1.12.1/jquery-ui.theme.mjin.css")
# print(dir(content))
# print(content.code)

import xml.dom.minidom

content = pyrest.webscrapping(r"https://stackoverflow.com/feeds/tag/tabulator")
dom_obj = xml.dom.minidom.parseString(content)

for entry in dom_obj.getElementsByTagName("entry"):
  t = entry.getElementsByTagName("title")[0]
  print(t._get_firstChild().nodeValue)




from epyk.core.html.graph import GraphDC


class DC(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "DC"

  def set_crossfilter(self, record, y_columns, x_axis, varName, extra_cols=None):
    """
    Description:
    -----------
    Set a crossfilter object and add the dimensions which will be added to a chart.

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    """
    #
    if not isinstance(y_columns, list):
      y_columns = [y_columns]

    if not isinstance(x_axis, tuple):
      x_axis = (x_axis, str)
    crossfilter = self.parent.context.rptObj.js.data.crossfilter(record, "%s_xf" % varName)
    if len(y_columns) == 1:
      dimension = crossfilter.dimension([x_axis] + (extra_cols or []), '%s_xf_dim' % varName)
      group = dimension.group('%s_xf_group' % varName).reduceSum(y_columns[0])
    else:
      dimension = crossfilter.dimension([x_axis] + (extra_cols or []) + [(y, int) for y in y_columns], '%s_xf_dim' % varName)
      group = dimension.group('%s_xf_group' % varName)

    #
    self.parent.context.rptObj._props.setdefault('js', {}).setdefault('datasets', {})["%s_xf" % varName] = crossfilter.toStr()
    self.parent.context.rptObj._props.setdefault('js', {}).setdefault('datasets', {})["%s_xf_dim" % varName] = dimension.toStr()
    self.parent.context.rptObj._props.setdefault('js', {}).setdefault('datasets', {})["%s_xf_group" % varName] = group.toStr()

    return {"crossfilter": crossfilter, 'dimension': dimension, 'group': group}

  def line(self, record=None, y_columns=None, x_axis=None, title=None, profile=None, options=None, width=(100, "%"),
           height=(330, "px"), htmlCode=None):
    """
    Description:
    -----------

    Related Pages:

			https://square.github.io/crossfilter/
    https://dc-js.github.io/dc.js/

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    if isinstance(y_columns, list):
      line_chart = self.series(record, y_columns, x_axis, 'line', title, profile, options, width, height, htmlCode)
    else:
      line_chart = GraphDC.ChartLine(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
      line_chart.dom.height(height[0]).x().yAxisLabel(y_columns).renderArea(True)
      if record is not None:
        self.set_crossfilter(record, y_columns, x_axis)
        cross_filter = self.set_crossfilter(record, y_columns, x_axis, line_chart.htmlId)
        line_chart.dom.dimension(cross_filter['dimension'].varId).group(cross_filter['group'].varId)
    self.parent.context.register(line_chart)
    return line_chart

  def series(self, record=None, y_columns=None, x_axis=None, series_type='line', title=None, profile=None, options=None, width=(100, "%"),
           height=(330, "px"), htmlCode=None):
    """
    Description:
    -----------

    Related Pages:

			https://square.github.io/crossfilter/
    https://dc-js.github.io/dc.js/

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    pivot_rec = []
    for rec in record:
      pivot_rec.extend([{'x': rec[x_axis], "name": "Series %s" % y, "y": rec[y]} for y in y_columns])

    line_chart = GraphDC.ChartSeries(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    cross_filter = self.set_crossfilter(pivot_rec, ["y"], "x", line_chart.htmlId, extra_cols=[('name', str)])
    #self.parent.context.rptObj._props.setdefault('js', {}).setdefault('datasets', {})['data_cf_%s' % line_id] = "var %(cId)s_cf = crossfilter(%(data)s); var %(cId)s_dim = %(cId)s_cf.dimension(function(d) {return [d.name, +d.x]})" % {'cId': line_id, 'data': pivot_rec}
    if series_type == 'line':
      line_chart.dom.line().height(height[0]).x().seriesAccessorByKey(1).keyAccessor(0).valueAccessor().elasticY(True) # .dimension("%s_dim" % line_id).group("%(cId)s_dim.group().reduceSum(function(d) {return d.y ;})" % {'cId': line_id})
      line_chart.dom.dimension(cross_filter['dimension'].varId).group(cross_filter['group'].varId)
    elif series_type == 'step':
      line_chart.dom.line().height(height[0]).x().seriesAccessorByKey(1).keyAccessor(0).valueAccessor().elasticY(True) # .dimension("%s_dim" % line_id).group("%(cId)s_dim.group().reduceSum(function(d) {return d.y ;})" % {'cId': line_id})
      line_chart.dom._sub_chart.renderArea(True).curveStepBefore()
      line_chart.dom.dimension(cross_filter['dimension'].varId).group(cross_filter['group'].varId)
    elif series_type == 'bubble':
      line_chart.dom.bubble().height(height[0]).x().seriesAccessorByKey(1).keyAccessor(0).valueAccessor().radiusValueAccessorByKey()
      line_chart.dom.dimension(cross_filter['dimension'].varId).group(cross_filter['group'].varId)
    elif series_type in ['scatter', 'point']:
      line_chart.dom.scatter().height(height[0]).x().seriesAccessorByKey(1).keyAccessor(0).valueAccessor()
      line_chart.dom.dimension(cross_filter['dimension'].varId).group(cross_filter['group'].varId)
    elif series_type == 'bar':
      line_chart.dom.bar().height(height[0]).x().seriesAccessorByKey(1).keyAccessor(0).valueAccessor()# .dimension("%s_dim" % line_id).group("%(cId)s_dim.group().reduceSum(function(d) {return d.y ;})" % {'cId': line_id})
      line_chart.dom.dimension(cross_filter['dimension'].varId).group(cross_filter['group'].varId)
    self.parent.context.register(line_chart)
    return line_chart

  def scatter(self, record=None, y_columns=None, x_axis=None, title=None, profile=None, options=None, width=(100, "%"),
           height=(330, "px"), htmlCode=None):
    """
    Description:
    -----------

    Related Pages:

			https://square.github.io/crossfilter/
    https://dc-js.github.io/dc.js/

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    if isinstance(y_columns, list):
      line_chart = self.series(record, y_columns, x_axis, 'scatter', title, profile, options, width, height, htmlCode)
    else:
      line_chart = GraphDC.ChartScatter(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
      line_chart.dom.height(height[0]).x().yAxisLabel(y_columns)
      if record is not None:
        cross_filter = self.set_crossfilter(record, y_columns, x_axis, line_chart.htmlId)
        line_chart.dom.dimension(cross_filter['dimension'].varId).group(cross_filter['group'].varId)
    self.parent.context.register(line_chart)
    return line_chart

  def step(self, record=None, y_columns=None, x_axis=None, title=None, profile=None, options=None, width=(100, "%"),
           height=(330, "px"), htmlCode=None):
    """
    Description:
    -----------

    Related Pages:
Attributes:
    ----------
    :param record:
    :param y_column:
    :param x_axis:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """

    if isinstance(y_columns, list):
      line_chart = self.series(record, y_columns, x_axis, 'step', title, profile, options, width, height, htmlCode)
    else:
      line_chart = GraphDC.ChartLine(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
      line_chart.dom.height(height[0]).x().yAxisLabel(y_columns).renderArea(True).curveStepBefore()
      if record is not None:
        cross_filter = self.set_crossfilter(record, y_columns, x_axis, line_chart.htmlId)
        line_chart.dom.dimension(cross_filter['dimension'].varId).group(cross_filter['group'].varId)
      self.parent.context.register(line_chart)
    return line_chart

  def bar(self, record=None, y_columns=None, x_axis=None, title=None, profile=None, options=None, width=(100, "%"),
           height=(330, "px"), htmlCode=None):
    """
    Description:
    -----------

    Related Pages:
Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    if isinstance(y_columns, list):
      bar_chart = self.series(record, y_columns, x_axis, 'bar', title, profile, options, width, height, htmlCode)
    else:
      bar_chart = GraphDC.ChartBar(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
      bar_chart.dom.height(height[0]).x().controlsUseVisibility(True).yAxisLabel(y_columns)
      if record is not None:
        cross_filter = self.set_crossfilter(record, y_columns, x_axis, bar_chart.htmlId)
        bar_chart.dom.dimension(cross_filter['dimension'].varId).group(cross_filter['group'].varId)
    self.parent.context.register(bar_chart)
    return bar_chart

  def hbar(self, record=None, y_column=None, x_axis=None, title=None, profile=None, options=None, width=(100, "%"),
           height=(330, "px"), htmlCode=None):
    """
    Description:
    -----------

    Related Pages:
Attributes:
    ----------
    :param record:
    :param y_column:
    :param x_axis:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    bar_chart = GraphDC.ChartRow(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    line_id = bar_chart.htmlId
    bar_chart.dom.height(height[0]).x().chartGroup(line_id).elasticX(True)
    if record is not None:
      cross_filter = self.set_crossfilter(record, y_column, x_axis, bar_chart.htmlId)
      bar_chart.dom.dimension(cross_filter['dimension'].varId).group(cross_filter['group'].varId)
    self.parent.context.register(bar_chart)
    return bar_chart

  def pie(self, record=None, y_column=None, x_axis=None, title=None, profile=None, options=None, width=(100, "%"),
           height=(330, "px"), htmlCode=None):
    """
    Description:
    -----------

    Related Pages:

			https://square.github.io/crossfilter/
    https://dc-js.github.io/dc.js/

    Attributes:
    ----------
    :param record:
    :param y_column:
    :param x_axis:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    pie_chart = GraphDC.ChartPie(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    pie_chart.dom.height(height[0])
    if record is not None:
      cross_filter = self.set_crossfilter(record, y_column, x_axis, pie_chart.htmlId)
      pie_chart.dom.dimension(cross_filter['dimension'].varId).group(cross_filter['group'].varId)
    self.parent.context.register(pie_chart)
    return pie_chart

  def sunburst(self, record=None, y_column=None, x_axis=None, title=None, profile=None, options=None, width=(100, "%"),
           height=(330, "px"), htmlCode=None):
    """

    Related Pages:

			https://square.github.io/crossfilter/
    https://dc-js.github.io/dc.js/

    :param data:
    :param seriesNames:
    :param xAxis:
    :param otherDims:
    :param title:
    :param profile:
    :param xAxisOrder:
    :param width:
    :param height:
    :param htmlCode:
    """
    pie_chart = GraphDC.ChartSunburst(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    line_id = pie_chart.htmlId
    self.parent.context.rptObj._props.setdefault('js', {}).setdefault('datasets', {})['data_cf_%s' % line_id] = "var %(cId)s_cf = crossfilter(%(data)s); var %(cId)s_dim = %(cId)s_cf.dimension(function(d) {return +d['%(x)s'];})" % {'cId': line_id, 'data': record, 'x': x_axis}
    pie_chart.dom.dimension("%s_dim" % line_id).group("%(cId)s_dim.group().reduceSum(function(d) {return d['%(y)s'] ;})" % {'cId': line_id, 'y': y_column})
    self.parent.context.register(pie_chart)
    return pie_chart

  def bubble(self, record=None, y_columns=None, x_axis=None, r_axis=None, title=None, profile=None, options=None, width=(100, "%"),
             height=(330, "px"), htmlCode=None):
    """
    Description:
    -----------

    Related Pages:

			https://square.github.io/crossfilter/
    https://dc-js.github.io/dc.js/
    https://www.tutorialspoint.com/dcjs/dcjs_bubble_chart.htm

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    options = options or {}
    if isinstance(y_columns, list):
      bubble_chart = self.series(record, y_columns, x_axis, 'bubble', title, profile, options, width, height, htmlCode)
    else:
      bubble_chart = GraphDC.ChartBubble(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
      bubble_chart.dom.height(height[0]).x().yAxisLabel(y_columns).keyAccessor(0).radiusValueAccessorByKey(1, statc_factor=options.get('statc_factor'))
      if record is not None:
        cross_filter = self.set_crossfilter(record, y_columns, x_axis, bubble_chart.htmlId, extra_cols=[(r_axis, int)])
        bubble_chart.dom.dimension(cross_filter['dimension'].varId).group(cross_filter['group'].varId)
    self.parent.context.register(bubble_chart)
    return bubble_chart




class Font(object):
  size, header_size, unit = 12, 14, "px"
  family = "Calibri"


class Icon(object):
  small, normal, big, unit = 10, 15, 20, 'px'


def font(step=0):
  """
  Description:
  ------------
  Relative change of the CSS font size based on the default one in the Font class.
  Changing the value Font.size will impact this function accordingly

  Usage::

    font(2)

  Attributes:
  ----------
  :param step: integer. The value to be added to the default font size
  """
  return "%s%s" % (Font.size+step, Font.unit)


def header(step=0):
  """
  Description:
  ------------

  Usage::

  Attributes:
  ----------
  :param step:
  """
  return "%s%s" % (Font.header_size+step, Font.unit)


def inline(cssAttrs):
  """
  Description:
  ------------
  Convert a CSS attributes dictionary to a online CSS Style to be added to the dom object

  Usage::

    inline({"color": "red"})

  Related Pages:
  --------------
  https://www.w3schools.com/css/css_howto.asp

  Attributes:
  ----------
  :param cssAttrs: Dictionary. The CSS Attributes
  """
  return ";".join(["%s: %s" % (k, v) for k, v in cssAttrs.items()])


def px_to_em(value, with_unit=True):
  """
  Description:
  ------------
  Convert the pixel value to em

  Related Pages:
  --------------
  https://www.w3schools.com/cssref/css_pxtoemconversion.asp

  Attributes:
  ----------
  :param value: Float. A pixel value
  :param with_unit: Boolean. To define the return format
  """
  em_value = value / 16
  if with_unit:
    return "%sem" % em_value

  return em_value


def em_to_px(value, with_unit=True):
  """
  Description:
  ------------
  Convert a em value in pixel

  Related Pages:
  --------------
  https://www.w3schools.com/cssref/css_pxtoemconversion.asp

  Attributes:
  ----------
  :param value: Float. The em value
  :param with_unit: Boolean. To define the return format
  """
  px_value = value * 16
  if with_unit:
    return "%spx" % px_value

  return px_value

#
DEFAULT_STYLE = "no_border"


# Default CSS Styles
BODY_CONTAINER = None # The body CSS dictionary
BACKGROUND = ('greys', 0)
MEDIA = 600


# Default CSS
CSS_EXCEPTIONS = True
CSS_EXCEPTIONS_FORMAT = "CSS - %s - invalid %s"




import json

from epyk.core.js.html import JsHtml
from epyk.core.js import JsUtils

from epyk.core.js.primitives import JsObjects


class Console(JsHtml.JsHtmlRich):

  def write(self, data, timestamp=None, stringify=False, skip_data_convert=False, format=None, profile=False):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data:
    :param timestamp:
    :param profile:
    :param stringify:
    :param skip_data_convert:
    :param format: String. A string output format using %s to define the data in the string
    """
    js_data = data if skip_data_convert else JsUtils.jsConvertData(data, None)
    if stringify:
      js_data = "JSON.stringify(%s)" % js_data
    if self._src.options.showdown is not False:
      js_data = '''
        (function(d){ var conv = new showdown.Converter(%s);
            let frag = document.createRange().createContextualFragment(conv.makeHtml(d)); 
            if((frag.firstChild === null) || (typeof frag.firstChild.style == "undefined")){return d}
            else{frag.firstChild.style.display = 'inline-block';frag.firstChild.style.margin = 0 ;  
                 return frag.firstChild.outerHTML}})(%s) ''' % (json.dumps(self._src.options.showdown), js_data)
    if format is not None:
      js_data = JsUtils.jsConvertData(format, None).toStr().replace("%s", '"+ %s +"') % js_data
    if timestamp or (self._src.options.timestamp and timestamp != False):
      return JsObjects.JsObjects.get("%s.innerHTML += ' > '+ new Date().toISOString().replace('T', ' ').slice(0, 19) +', '+ %s +'<br/>'" % (self.varName, js_data))

    return JsObjects.JsObjects.get("%s.innerHTML += ' > '+ %s +'<br/>'" % (self._src.dom.varId, js_data))

  def clear(self):
    """

    :return:
    """
    return JsObjects.JsObjects.get('%s.innerHTML = ""' % self.varName)


class Editor(JsHtml.JsHtmlRich):

  def timestamp(self):
    """
    Description:
    ------------
    Update the editor timestamp
    """
    return self.querySelector("span").innerHTML(JsObjects.JsDate.JsDate().getStrTimeStamp())


class CodeMirror(JsHtml.JsHtmlRich):

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.htmlId = varName if varName is not None else htmlObj.htmlId
    self.varName, self.varData, self.__var_def = "%s.getWrapperElement()" % htmlObj.editorId, "", None
    self._src, self._report = htmlObj, report
    self._js = []
    self._jquery, self._jquery_ui, self._d3 = None, None, None

  @property
  def content(self):
    """
    Description:
    -----------

    :return:
    """
    return JsHtml.ContentFormatters(self._report, "%s.getValue()" % self._src.editorId)

  def select(self):
    """
    Description:
    -----------
    Select the whole content of the editor.

    Related Pages:

			https://codemirror.net/3/doc/manual.html#keymaps
    """
    return JsObjects.JsObjects.get("%s.execCommand('selectAll')" % self._src.editorId)

  def singleSelection(self):
    """
    Description:
    -----------
    When multiple selections are present, this deselects all but the primary selection.

    Related Pages:

			https://codemirror.net/3/doc/manual.html#keymaps
    """
    return JsObjects.JsObjects.get("%s.execCommand('singleSelection')" % self._src.editorId)

  def killLine(self):
    """
    Description:
    -----------
    Emacs-style line killing. Deletes the part of the line after the cursor.
    If that consists only of whitespace, the newline at the end of the line is also deleted.

    Related Pages:

			https://codemirror.net/3/doc/manual.html#keymaps
    """
    return JsObjects.JsObjects.get("%s.execCommand('killLine')" % self._src.editorId)

  def deleteLine(self):
    """
    Description:
    -----------
    Deletes the whole line under the cursor, including newline at the end.

    Related Pages:

			https://codemirror.net/3/doc/manual.html#keymaps
    """
    return JsObjects.JsObjects.get("%s.execCommand('deleteLine')" % self._src.editorId)

  def copy(self):
    """
    Description:
    -----------
    """
    return JsObjects.JsObjects.get("%s.execCommand('copy')" % self._src.editorId)

  def setValue(self, data):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param data:

    :return:
    """
    data = JsUtils.jsConvertData(data, None)
    return JsObjects.JsObjects.get("%s.setValue(%s)" % (self._src.editorId, data))

  def setOption(self, name, value):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param name:
    :param value:

    :return:
    """
    name = JsUtils.jsConvertData(name, None)
    value = JsUtils.jsConvertData(value, None)
    return JsObjects.JsObjects.get("%s.setOption(%s, %s)" % (self._src.editorId, name, value))

  def refresh(self):
    """
    Description:
    -----------

    :return:
    """
    return JsObjects.JsObjects.get("%s.refresh()" % self._src.editorId)

  def clear(self):
    """
    Description:
    -----------

    :return:
    """
    return JsObjects.JsObjects.get('%s.setValue("")' % self._src.editorId)

  def empty(self):
    """
    Description:
    -----------

    :return:
    """
    return JsObjects.JsObjects.get('%s.setValue("")' % self._src.editorId)



from epyk.core.js.fncs import JsFncsRecords
from epyk.core.js.fncs import JsFncsAgg

from epyk.core.js.objects import JsChartNvd3
from epyk.core.js.objects import JsChartsJs
from epyk.core.js.objects import JsChartD3
from epyk.core.js.objects import JsChartBillboard
from epyk.core.js.objects import JsChartPlotly

from epyk.core.js.primitives import JsObject

from epyk.core.js import JsUtils


class FncToObject(object):
  def __init__(self, data, js_src, data_schema=None):
    """
    :param data:
    :param js_src:
    :param data_schema:
    """
    self._js_src, self._data_schema, self._data = js_src, data_schema, data

  def __register_records_fnc(self, fnc_name, fnc_def, fnc_pmts=None):
    """
    This function will attach to the report object only the javascript functions used during the report

    :param fnc_name:
    :param fnc_def:

    :return:
    """
    fnc_pmts = ["data"] + (fnc_pmts or [])
    if not fnc_name in self._js_src.get('js', {}).get('functions', {}):
      self._js_src.setdefault('js', {}).setdefault('functions', {})[fnc_name] = {'content': "var result = []; %s;return result" % JsUtils.cleanFncs(fnc_def), 'pmt': fnc_pmts}

  @property
  def chartJs(self):
    """
    Data transformation to the ChartJs package
    """
    return JsChartsJs.JsChartLinks(self._data, self._js_src, self._data_schema)

  @property
  def nvd3(self):
    """
    Data transformation to the NVD3 package
    """
    return JsChartNvd3.JsNVD3Links(self._data, self._js_src, self._data_schema)

  @property
  def c3(self):
    """
    Data transformation to the C3 package
    """
    return JsChartBillboard.JsChartBillboardLinks(self._data, self._js_src, self._data_schema)

  @property
  def billboard(self):
    """
    Data transformation to the Billboard package
    """
    return JsChartBillboard.JsChartBillboardLinks(self._data, self._js_src, self._data_schema)

  @property
  def d3(self):
    """
    Data transformation to the D3 package
    """
    return JsChartD3.JsChartD3Links(self._data, self._js_src, self._data_schema)

  @property
  def dc(self):
    """
    Data transformation to the DC package
    """
    from epyk.core.js.objects import JsChartDC

    return JsChartDC.JsChartDCLinks(self._data, self._js_src, self._data_schema)

  @property
  def plotly(self):
    """
    Data transformation to the Plotly package
    """
    return JsChartPlotly.JsChartPlotlyLinks(self._data, self._js_src, self._data_schema)


class FncRoAggRec(object):
  def __init__(self, data, js_src, data_schema=None):
    self._js_src, self._data_schema, self._data = js_src, data_schema, data

  def __register_records_fnc(self, fnc_name, fnc_def, fnc_pmts=None):
    """
    This function will attach to the report object only the javascript functions used during the report

    :param fnc_name:
    :param fnc_def:

    :return:
    """
    fnc_pmts = ["data"] + (fnc_pmts or [])
    if not fnc_name in self._js_src.get('js', {}).get('functions', {}):
      self._js_src.setdefault('js', {}).setdefault('functions', {})[fnc_name] = {'content': "var result = []; %s;return result" % JsUtils.cleanFncs(fnc_def), 'pmt': fnc_pmts}

  def sum(self, column):
    """
    Get the result only on the selected column. This function will return a dictionary with the column name and the sum

    :param column: String. The column name in the records
    :return: The data object
    """
    fnc_name = JsFncsAgg.JsAggColStats.__name__
    self.__register_records_fnc(fnc_name, JsFncsAgg.JsAggColStats.content, fnc_pmts=list(JsFncsAgg.JsAggColStats.pmts))
    column = JsUtils.jsConvertData(column, None)
    self._data_schema['fncs'].append("%s(%%s, %s)" % (fnc_name, column))
    return self._data

  def sum_with_kpi(self, column):
    """
    Return the aggregated value of a defined column with some KPI (count, average, max, min)
    Get the result only on the selected column. This function will return a dictionary with the column name and the sum

    :param column: String. The column name in the records
    :return: The data object
    """
    fnc_name = JsFncsAgg.JsAggColStats.__name__
    self.__register_records_fnc(fnc_name, JsFncsAgg.JsAggColStats.content, fnc_pmts=list(JsFncsAgg.JsAggColStats.pmts))
    column = JsUtils.jsConvertData(column, None)
    self._data_schema['fncs'].append("%s(%%s, %s)" % (fnc_name, column))
    return self._data

  def max(self, column):
    """

    :param column: String. The column name in the records
    :return: The data object
    """
    fnc_name = JsFncsAgg.JsAggColMax.__name__
    self.__register_records_fnc(fnc_name, JsFncsAgg.JsAggColMax.content, fnc_pmts=list(JsFncsAgg.JsAggColMax.pmts))
    column = JsUtils.jsConvertData(column, None)
    self._data_schema['fncs'].append("%s(%%s, %s)" % (fnc_name, column))
    return self._data

  def min(self, column):
    """

    :param column: String. The column name in the records
    :return: The data object
    """
    fnc_name = JsFncsAgg.JsAggColMin.__name__
    self.__register_records_fnc(fnc_name, JsFncsAgg.JsAggColMin.content, fnc_pmts=list(JsFncsAgg.JsAggColMin.pmts))
    column = JsUtils.jsConvertData(column, None)
    self._data_schema['fncs'].append("%s(%%s, %s)" % (fnc_name, column))
    return self._data

  def eq(self, column, val):
    """
    Return the last record in the records set matching the condition

    :param column: String. The column name in the records
    :param val: Object. The corresponding value

    :return: The data object
    """
    fnc_name = JsFncsAgg.JsAggColEq.__name__
    self.__register_records_fnc(fnc_name, JsFncsAgg.JsAggColEq.content, fnc_pmts=list(JsFncsAgg.JsAggColEq.pmts))
    column = JsUtils.jsConvertData(column, None)
    val = JsUtils.jsConvertData(val, None)
    self._data_schema['fncs'].append("%s(%%s, %s, %s)" % (fnc_name, column, val))
    return self._data


class FncOnRecords(object):
  def __init__(self, data, js_src, data_schema=None, profile=False):
    self._js_src, self._data_schema, self._data = js_src, data_schema, data

  @property
  def o(self):
    """
    Property to the data final object.
    Those items help to the link to external packages
    """
    return FncToObject(self._js_src, self._data_schema)

  def __register_records_fnc(self, fnc_name, fnc_def, fnc_pmts=None, profile=False):
    """
    This function will attach to the report object only the javascript functions used during the report

    :param fnc_name:
    :param fnc_def:
    :param fnc_pmts:

    :return:
    """
    fnc_pmts = ["data"] + (fnc_pmts or [])
    if not fnc_name in self._js_src.get('js', {}).get('functions', {}):
      self._js_src.setdefault('js', {}).setdefault('functions', {})[fnc_name] = {'content': "var result = []; %s;return result" % JsUtils.cleanFncs(fnc_def), 'pmt': fnc_pmts}

  def custom(self, fnc_name, fnc_content, fnc_pmts=None, profile=False):
    """

    The function content should use data and produce an object record

    :param fnc_name: A string for the Javascript function name
    :param fnc_content: The javascript function content
    :param fnc_pmts: A String, The Javascript function parameters
    :param profile: A Boolean flag to activate the profiling

    :return: "This" in order to allow the chains
    """
    self.__register_records_fnc(fnc_name, fnc_content, fnc_pmts, profile)
    return self

  def url(self):
    """

    :return:
    """
    fnc_name = JsFncsRecords.JsToUrl.__name__
    fnc_pmts = ["data"]
    for p in getattr(JsFncsRecords.JsToUrl, 'params', []):
      fnc_pmts.append(p)
    if not fnc_name in self._js_src.get('functions', {}):
      content = JsUtils.cleanFncs(JsFncsRecords.JsToUrl.value)
      self._js_src.setdefault('functions', {})[fnc_name] = {'content': "%s; return result" % content, 'pmt': fnc_pmts}
    return fnc_name

  def count(self, keys, values=None, profile=False):
    """

    The Javascript function are using the main data as a first parameter

    If values is defined, the Javascript will aggregate the data based on the composite key and the values will be
    available in the record. Also the count will be displayed.
    The values will be one in the record and not the sum

    :param keys: A list with the column names
    :param values: A list with the values to keep in the result record

    :return: "This" to allow function chains
    """
    if not isinstance(keys, list):
      keys = [keys]

    if values is None:
      fnc_name = JsFncsRecords.JsCountAll.__name__
      self.__register_records_fnc(fnc_name, JsFncsRecords.JsCountAll.value, fnc_pmts=list(JsFncsRecords.JsCountAll.params))
      self._data_schema['fncs'].append("%s(%%s, %s)" % (fnc_name, keys))
    else:
      fnc_name = JsFncsRecords.JsCount.__name__
      self.__register_records_fnc(fnc_name, JsFncsRecords.JsCount.value, fnc_pmts=list(JsFncsRecords.JsCount.params))
      self._data_schema['fncs'].append("%s(%%s, %s, %s)" % (fnc_name, keys, values))
    return self._data

  def count_with_kpi(self, keys, values, profile=False):
    """

    :param keys:
    :param values:
    :param profile:
    :return:
    """
    fnc_name = JsFncsRecords.JsCountSum.__name__
    self.__register_records_fnc(fnc_name, JsFncsRecords.JsCountSum.value, fnc_pmts=list(JsFncsRecords.JsCountSum.params))
    self._data_schema['fncs'].append("%s(%%s, %s, %s)" % (fnc_name, keys, values))
    return self._data

  def count_distinct(self, keys, profile=False):
    """

    :param keys:

    :return: "This" to allow function chains
    """
    if not isinstance(keys, list):
      keys = [keys]
    fnc_name = JsFncsRecords.JsCountDistinct.__name__
    self.__register_records_fnc(fnc_name, JsFncsRecords.JsCountDistinct.value, fnc_pmts=list(JsFncsRecords.JsCountDistinct.params))
    self._data_schema['fncs'].append("%s(%%s, %s)" % (fnc_name, keys))
    return self._data

  def top(self, column, n=1, order='desc', profile=False):
    """
    The Javascript function are using the main data as a first parameter

    Example

    :param column:
    :param n:
    :param order:

    :return: "This" to allow function chains
    """
    fnc_name = JsFncsRecords.JsTop.__name__
    self.__register_records_fnc(fnc_name, JsFncsRecords.JsTop.value, fnc_pmts=list(JsFncsRecords.JsTop.params))
    self._data_schema['fncs'].append("%s(%%s, %s, '%s', '%s')" % (fnc_name, n, column, order))
    return self


class FncFiltere(object):
  def __init__(self, data, js_src, data_schema=None, profile=False):
    self._js_src, self._data_schema, self._data = js_src, data_schema, data
    fnc_name = JsFncsRecords.JsFilter.__name__
    fnc_pmts = ["data"] + (list(JsFncsRecords.JsFilter.pmts) or [])
    if not fnc_name in self._js_src.get('js', {}).get('functions', {}):
      self._js_src.setdefault('js', {}).setdefault('functions', {})[fnc_name] = {
        'content': "var result = []; %s;return result" % JsUtils.cleanFncs(JsFncsRecords.JsFilter.content), 'pmt': fnc_pmts}
    self._data_schema['filters'] = []

  def custom(self, column, val, compare_type, all_if_empty=True):
    filter_data = JsUtils.jsConvertData({"colName": column, "val": val, "op": compare_type, "allIfEmpty": all_if_empty}, None)
    self._data_schema['filters'].append(filter_data.toStr())
    return self._data

  def not_in_(self):
    pass

  def not_range_(self, column, val, compare_type="in", all_if_empty=True):
    pass

  def in_(self, column, val):
    """

    :param column:
    :param val:
    """
    return self.custom(column, val, "in", True)

  def range_(self, column, val, strict_left=False, strict_right=False):
    """

    :param column:
    :param val:
    :param strict_left:
    :param strict_right:
    """
    if not strict_left:
      if not strict_right:
        return self.custom(column, val, "><=", True)

      return self.custom(column, val, "><", True)

    if not strict_right:
      if not strict_left:
        return self.custom(column, val, "=><", True)

      return self.custom(column, val, "><", True)

    return self.custom(column, val, "=><=", True)

  def eq_(self, column, val):
    """

    :param column:
    :param val:
    """
    return self.custom(column, val, ">", True)

  def sup_(self, column, val, strict=False):
    """
    Filter only the data above the value for the given key in the record

    :param column: String. The column name
    :param val: Object. The value in the dictionary
    :param strict: Boolean. A flag to specify if the value should be included
    """
    if strict:
      return self.custom(column, val, ">", True)

    return self.custom(column, val, ">=", True)

  def inf_(self, column, val, strict=False):
    """
    Filter only the data below the value for the given key in the record

    :param column: String. The column name
    :param val: Object. The value in the dictionary
    :param strict: Boolean. A flag to specify if the value should be included
    """
    if strict:
      return self.custom(column, val, "<", True)

    return self.custom(column, val, "<=", True)


class JsRegisteredFunctions(object):
  class __internal(object):
    _props = {}

  def __init__(self, src=None):
    src = src or self.__internal()
    if not 'js' in src._props:
      src._props['js'] = {}
    self._js_src = src._props['js']

  def cssStyle(self, params):
    """

    :return:
    """
    self._js_src.setdefault('functions', {})["cssStyle"] = {
      'content': 'cssParams = []; for(var i in params){cssParams.push(i +":"+ params[i])}; return cssParams.join(";")',
      'pmt': ["params"]}
    return "cssStyle"

  def anonymous(self, jsFnc, pmts=None):
    """
    Create a anonymous / lambda function.
    Those functions are directly called when they are defined.

    Related Pages:

			https://www.w3schools.com/js/js_function_definition.asp

    :param jsFnc:
    :param pmts:

    :return:
    """
    if pmts is None:
      return JsFunction("(function(){%s})()" % jsFnc)

    return JsFunction("(function(%s){%s})()" % (",".join(pmts), jsFnc))

  def get(self, fnc_name, *args):
    """
    Call a bespoke functions on the Javascript side

    :param fnc_name: The function name
    :param args: The different arguments in the function definition

    :return: The Javascript sting
    """
    pmts = [str(JsUtils.jsConvertData(p, None)) for p in args]
    return "%s(%s)" % (fnc_name, ", ".join(pmts))

  def inline(self, fnc_name, jsFnc, pmts=None):
    """
    Create a name function which can be then called later

    Related Pages:

			https://www.w3schools.com/js/js_function_definition.asp

    :param fnc_name:
    :param jsFnc:
    :param pmts:

    :return: The function name which can be used in the Javascript
    """
    self._js_src.setdefault('functions', {})[fnc_name] = {'content': JsUtils.jsConvertFncs(jsFnc, toStr=True), 'pmt': pmts}
    return fnc_name

  @property
  def records(self):
    """
    Javascript pre defined function dedicated to transform a records.
    Namely a list of dictionaries
    """
    return FncOnRecords(self._js_src)


class JsFunction(object):
  """

  """
  fncName = "lambda"

  def __init__(self, strFnc):
    self.__strFnc = strFnc

  def __str__(self):
    return self.__strFnc

  def toStr(self):
    return self.__strFnc


class JsFunctions(list):
  """

  """
  def __init__(self, strFnc):
    if not isinstance(strFnc, list):
      strFnc = [strFnc]
    self.__strFncs = strFnc

  def append(self, strFnc):
    self.__strFncs.append(strFnc)

  def extend(self, strFncs):
    self.__strFncs.extend(strFncs)

  def toStr(self):
    return "; ".join([s.toStr() if hasattr(s, 'toStr') else str(s) for s in self.__strFncs])


_JSFNCS = 0


class JsLambda(object):
  """

  """

  def __init__(self):
    """
    """
    global _JSFNCS

    _JSFNCS += 1
    self.fncName = "function_%s" % _JSFNCS


class JsTypeOf(object):
  """

  """
  fncName = "typeof"

  def __init__(self, jsData):
    if self.fncName is None:
      raise Exception("Private fncName variable should be defined for pre defined functions ")

    self.__jsArgs = [jsData]

  def __str__(self):
    """

    :return:
    """
    return "%s(%s)" % (self.fncName, ", ".join([str(a) for a in self.__jsArgs]))


class JsAnonymous(object):

  def __init__(self, jsFncs):
    self.__strFnc, self.__returnFnc, self.__paramsFnc = jsFncs, "", []

  def return_(self, value):
    self.__returnFnc = value
    return self

  def params(self, pmts):
    """
    """
    self.__paramsFnc = pmts
    return self

  def call(self, *args, **kwargs):
    """

    :param args:
    :param kwargs:
    """
    _args = []
    if self.__paramsFnc:
      for a in list(args):
        _args.append(str(a))
      if kwargs:
        for i, p in enumerate(self.__paramsFnc):
          if p in kwargs:
            _args.append(str(kwargs[p]))
    return JsObject.JsObject("%s(%s)" % (self, ", ".join(_args)))

  def __str__(self):
    return "(function (%s) {%s; return %s})" % (", ".join(self.__paramsFnc), self.__strFnc, self.__returnFnc)

  def toStr(self):
    return str(self.__strFnc)




from epyk.core.js import JsUtils
from epyk.core.js.packages import JsPackage
from epyk.core.js.primitives import JsObjects


class Timepicker(JsPackage):

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.htmlId = varName if varName is not None else htmlObj.htmlId
    self.varName, self.varData, self.__var_def = "document.getElementById('%s')" % self.htmlId, "", None
    self._src, self._report = htmlObj, report
    self._js, self._jquery = [], None

  def change(self, jsFncs):
    """
    Description:
    ------------
    Event triggerd when the value of the input field changes. A Date object containing the selected time is passed as the first argument of the callback.

    Related Pages:

			https://timepicker.co/options/
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    jsFncs = JsUtils.jsConvertFncs(jsFncs, toStr=True)
    return JsObjects.JsObjects.get('%s.timepicker({change: function(time) { %s } })' % (self._src.dom.jquery.varId, jsFncs))

  def value(self, jsValue=None):
    """
    Description:
    ------------
    Set the timepicker object value

    Related Pages:

			https://stackoverflow.com/questions/32378842/setting-the-hour-and-minute-of-timepicker-dynamically

    Attributes:
    ----------
    :param jsValue: String or Js Object
    """
    jsValue = JsUtils.jsConvertData(jsValue, None)
    return JsObjects.JsObjects.get('%s.timepicker("setTime", %s)' % (self._src.dom.jquery.varId, jsValue))



"""
Javascript Aggregators

Those functions will return only 1 record within the recordset.
They are all returning list of records in order to allow function chains
"""


class JsAggColSum(object):
  pmts = ("column", )
  content = '''
    var value = 0;
    data.forEach(function(rec){if(rec[column] !== undefined){value += rec[column]}});
    result = [{label: column, value: value}]
    '''


class JsAggColMax(object):
  pmts = ("column", )
  content = '''
    var max_val = undefined;
    data.forEach(function(rec){if(rec[column] !== undefined){
      if((max_val == undefined) || (rec[column] > max_val[column])){max_val = rec}}});
    result = [max_val]
    '''


class JsAggColEq(object):
  pmts = ("column", "value")
  content = '''data.forEach(function(rec){if(rec[column] == value){result = [rec]}})'''


class JsAggColMin(object):
  pmts = ("column", )
  content = '''
    var min_val = undefined;
    data.forEach(function(rec){if(rec[column] !== undefined){
      if((min_val == undefined) || (rec[column] < min_val[column])){min_val = rec} }});
    result = [min_val]
    '''


class JsAggColStats(object):
  pmts = ("column", )
  content = '''
    var value = 0; var countVals = 0; var min_val = null; var max_val = null;
    data.forEach(function(rec){
      if(rec[column] !== undefined){value += rec[column]; countVals ++;
      if((min_val == null) || (rec[column] < min_val)){min_val = rec[column]};
      if((max_val == null) || (rec[column] > max_val)){max_val = rec[column]}}});
    result = [{label: column, value: value, count: countVals, average: value/countVals, max: max_val, min: min_val}]
    '''


